using System;
using System.Collections.Generic;
using System.Text;
using MrHomer.Database;
using MrHomer.Database.Query;

namespace DuffBeer.Database
{
    public class DataTable
    {
        private IDatabase myDB;
        private IQueryBuilder myQB;
        private string myTable, myPKey;
        private DBResults table;
        private int currentRow;
        private bool rowChanged, newRow;

        public enum SortOrder
        {
            Ascending,
            Descending
        }

        /// <summary>
        /// Creates a new Data Table
        /// </summary>
        /// <param name="source">The database storing this table</param>
        /// <param name="qb">The Query Builder to use</param>
        /// <param name="Table">The table to which this is bound</param>
        /// <param name="PKey">The primary key column for the table</param>
        public DataTable(IDatabase source, IQueryBuilder qb, string Table, string PKey)
        {
            myDB = source;
            myQB = qb;
            myTable = Table;
            myPKey = PKey;

            Refresh();
        }

        /// <summary>
        /// Gets the current row
        /// </summary>
        public Dictionary<string, string> Row
        {
            get { return table[currentRow]; }
            set { table[currentRow] = value; rowChanged = true; }
        }

        public int Count
        {
            get { return table.Length(); }
        }

        /// <summary>
        /// Gets the latest data from the database
        /// </summary>
        public void Refresh()
        {
            table = myDB.Query(myQB.Select(myTable) + ";");
            rowChanged = false;
            newRow = false;
        }

        /// <summary>
        /// Moves to the first record in the table
        /// </summary>
        public void MoveFirst()
        {
            if (rowChanged)
            {
                SaveRow();
                rowChanged = false;
                newRow = false;
            }
            currentRow = 0;
        }

        /// <summary>
        /// Moves to the last record in the table
        /// </summary>
        public void MoveLast()
        {
            if (rowChanged)
            {
                SaveRow();
                rowChanged = false;
                newRow = false;
            }
            currentRow = table.Length() - 1;
        }

        /// <summary>
        /// Tries to move to the next record in the table
        /// </summary>
        /// <returns><code>True</code> if able to advance, <code>False</code> if not</returns>
        public bool MoveNext()
        {
            if (rowChanged || newRow)
            {
                SaveRow();
                rowChanged = false;
                newRow = false;
            }
            if (currentRow < table.Length() - 1)
            {
                currentRow++;
                return true;
            }
            else
                return false;
        }

        /// <summary>
        /// Tries to move to the previous record in the table
        /// </summary>
        /// <returns><code>True</code> if able to rewind, <code>False</code> if not</returns>
        public bool MovePrev()
        {
            if (rowChanged || newRow)
            {
                SaveRow();
                rowChanged = false;
                newRow = false;
            }
            if (currentRow > 0)
            {
                currentRow--;
                return true;
            }
            else
                return false;
        }

        /// <summary>
        /// Sorts the table on the given column
        /// </summary>
        /// <param name="Column">The column to sort</param>
        /// <param name="Order">The sort order</param>
        public void Sort(string Column, SortOrder Order)
        {
            table = myDB.Query(myQB.OrderBy(myQB.Select(myTable), new KeyValuePair<string,bool>(Column, Order == SortOrder.Descending)) + ";");
            rowChanged = false;
            newRow = false;
        }

        /// <summary>
        /// Searches through the table for the given condition
        /// </summary>
        /// <param name="column">The relevant column</param>
        /// <param name="value">The value to find</param>
        /// <returns><code>True</code> if found, <code>False</code> if not</returns>
        public bool Find(string column, string value)
        {
            if (rowChanged || newRow)
            {
                SaveRow();
                rowChanged = false;
                newRow = false;
            }

            int where = currentRow;
            MoveFirst();
            do
            {
                if (Row == null) break;
                if (Row[column].ToString() == value)
                    return true;
            } while (MoveNext());

            // Restore previous position and fail.
            currentRow = where;
            return false;
        }

        /// <summary>
        /// Saves the current row to the database
        /// </summary>
        public void SaveRow()
        {
            List<string> keys = new List<string>(Row.Keys);
            List<string> values = new List<string>(Row.Values);
            List<KeyValuePair<string, string>> data = new List<KeyValuePair<string, string>>();

            for (int i = 0; i < keys.Count; i++)
            {
                if(keys[i] != myPKey)
                    data.Add(new KeyValuePair<string, string>(keys[i], values[i].ToString()));
            }

            // The result is unimportant.
            if (newRow)
            {
                myDB.Query(myQB.Insert(myTable, data.ToArray()));
                newRow = false;
                rowChanged = false;
            }
            else
            {
                myDB.Query(myQB.Update(myTable, " " + myPKey + "='" + Row[myPKey] + "' ", data.ToArray()));
                rowChanged = false;
            }
        }

        /// <summary>
        /// Adds a new row to the table and moves to it
        /// </summary>
        public void AddRow()
        {
            table.AddRow();
            MoveLast();
            newRow = true;
        }

        /// <summary>
        /// Deletes the current row from the database and moves to the next row
        /// </summary>
        /// <returns><code>True</code> if the pointer was advanced, <code>False</code> if it was rewound</returns>
        public bool DeleteRow()
        {
            myDB.Query(myQB.Delete(myTable, " " + myPKey + "='" + Row[myPKey] + "' "));
            Refresh();

            if (currentRow < table.Length() - 1)
                return MoveNext();
            else
                MoveLast();
            return false;
        }
    }
}
