using System;
using System.Collections.Generic;
using System.Text;
using MrHomer.Plugins;

namespace MrHomer.Database.Query
{
    /// <summary>
    /// Makes SQL queries
    /// </summary>
    [MrHomerPlugin("Type", "SQL")]
    public class QueryBuilder : IQueryBuilder
    {
        /// <summary>
        /// Makes a list of columns in SQL format
        /// </summary>
        /// <param name="Columns">The columns</param>
        /// <returns>A SQL-Formatted column lists</returns>
        private string ColumnString(params string[] Columns)
        {
            if (Columns.Length == 0) return "*";
            string res = Columns[0];
            for (int i = 1; i < Columns.Length; i++)
                res += ", " + Columns[i];

            return res;
        }

        /// <summary>
        /// Makes a list of columns in SQL format
        /// </summary>
        /// <param name="Columns">The columns</param>
        /// <returns>A SQL-Formatted column lists</returns>
        private string ColumnString(string Column1, params string[] Columns)
        {
            string res = Column1;
            for (int i = 0; i < Columns.Length; i++)
                res += ", " + Columns[i];

            return res;
        }

        /// <summary>
        /// Makes a SELECT query
        /// </summary>
        /// <param name="Table">The table</param>
        /// <param name="Columns">Collumns to return in the selection</param>
        /// <returns>A SQL Select Query</returns>
        public string Select(string Table, params string[] Columns)
        {
            return "SELECT " + ColumnString(Columns) + " FROM " + Table;
        }

        /// <summary>
        /// Makes a SELECT DISTINCT query
        /// </summary>
        /// <param name="Table">The table</param>
        /// <param name="Columns">Collumns to return in the selection</param>
        /// <returns>A SQL Select Query</returns>
        public string SelectDistinct(string Table, params string[] Columns)
        {
            return "SELECT DISTINCT " + ColumnString(Columns) + " FROM " + Table;
        }

        /// <summary>
        /// Makes a SELECT ... WHERE query
        /// </summary>
        /// <param name="Table">The table</param>
        /// <param name="Condition">The selection condition</param>
        /// <param name="Columns">Collumns to return in the selection</param>
        /// <returns>A SQL Select Query</returns>
        public string SelectWhere(string Table, string Condition, params string[] Columns)
        {
            return "SELECT " + ColumnString(Columns) + " FROM " + Table + " WHERE " + Condition;
        }

        /// <summary>
        /// Makes a SELECT DISTINCT ... WHERE query
        /// </summary>
        /// <param name="Table">The table</param>
        /// <param name="Condition">The selection condition</param>
        /// <param name="Columns">Collumns to return in the selection</param>
        /// <returns>A SQL Select Query</returns>
        public string SelectDistinctWhere(string Table, string Condition, params string[] Columns)
        {
            return "SELECT DISTINCT " + ColumnString(Columns) + " FROM " + Table + " WHERE " + Condition;
        }

        /// <summary>
        /// Makes an INSERT query
        /// </summary>
        /// <param name="Table">The table</param>
        /// <param name="values">A list of values to insert</param>
        /// <returns>The Query</returns>
        public string Insert(string Table, params KeyValuePair<string, string>[] values)
        {
            string res = "INSERT INTO " + Table + " (";
            if (values.Length > 0)
                res += values[0].Key;
            for (int i = 1; i < values.Length; i++)
                res += ", " + values[i].Key;
            res += ") VALUES ('";
            if (values.Length > 0)
                res += values[0].Value;
            for (int i = 1; i < values.Length; i++)
                res += "', '" + values[i].Value;
            res += "');";

            return res;
        }

        /// <summary>
        /// Makes an update query
        /// </summary>
        /// <param name="Table">The table to update</param>
        /// <param name="Condition">The update condition</param>
        /// <param name="values">A list of values to update</param>
        /// <returns>The Query</returns>
        public string Update(string Table, string Condition, params KeyValuePair<string, string>[] values)
        {
            string res = "UPDATE " + Table + " SET ";
            if (values.Length > 0)
                res += values[0].Key + "='" + values[0].Value + "'";
            for (int i = 1; i < values.Length; i++)
                res += ", " + values[i].Key + "='" + values[i].Value + "'";
            if (Condition != "")
                res += " WHERE " + Condition;
            res += ";";
            
            return res;
        }

        /// <summary>
        /// Makes a delete query
        /// </summary>
        /// <param name="Table">The table from which to delete</param>
        /// <param name="Condition">The delete condition</param>
        /// <returns>The Query</returns>
        public string Delete(string Table, string Condition)
        {
            return "DELETE FROM " + Table + " WHERE " + Condition + ";";
        }

        /// <summary>
        /// ANDs two conditions together
        /// </summary>
        /// <param name="Condition1">The first condition</param>
        /// <param name="Conditions">Additional conditions</param>
        /// <returns>The Query</returns>
        public string And(string Condition1, params string[] Conditions)
        {
            string res = Condition1;
            for (int i = 0; i < Conditions.Length; i++)
                res += " AND " + Conditions[i];

            return res + " ";
        }

        /// <summary>
        /// ORs two conditions together
        /// </summary>
        /// <param name="Condition1">The first condition</param>
        /// <param name="Conditions">Additional conditions</param>
        /// <returns>The Query</returns>
        public string Or(string Condition1, params string[] Conditions)
        {
            string res = Condition1;
            for (int i = 0; i < Conditions.Length; i++)
                res += " OR " + Conditions[i];

            return res + " ";
        }

        /// <summary>
        /// NOTs a condition
        /// </summary>
        /// <param name="Condition">The Conditon</param>
        /// <returns>The Query</returns>
        public string Not(string Condition)
        {
            return "NOT(" + Condition + ") ";
        }

        /// <summary>
        /// Makes an IN condition
        /// </summary>
        /// <param name="Column">The affected column</param>
        /// <param name="Option1">The first option</param>
        /// <param name="Options">Additional options</param>
        /// <returns>The Query</returns>
        public string In(string Column, string Option1, params string[] Options)
        {
            string res = Column + " IN ('" + Option1;
            for (int i = 0; i < Options.Length; i++)
                res += "', '" + Options[i];
            res += "') ";

            return res;
        }

        /// <summary>
        /// Makes a BETWEEN condition
        /// </summary>
        /// <param name="Column">The affected column</param>
        /// <param name="LBound">The lower boundary</param>
        /// <param name="UBound">The upper boundary</param>
        /// <returns>The Query</returns>
        public string Between(string Column, string LBound, params string[] UBound)
        {
            return Column + " BETWEEN '" + LBound + "' AND '" + UBound + "' ";
        }

        /// <summary>
        /// Makes a LIKE condition
        /// </summary>
        /// <param name="Column">The affected column</param>
        /// <param name="Condition">The condition</param>
        /// <returns>The Query</returns>
        public string Like(string Column, string Condition)
        {
            return Column + " LIKE '" + Condition + "' ";
        }

        /// <summary>
        /// Makes an ORDER BY clause
        /// </summary>
        /// <param name="Query">The query to sort</param>
        /// <param name="Column1">The first column to sort on</param>
        /// <param name="Columns">Additional columns to sort on</param>
        /// <returns>The Query</returns>
        /// <remarks>KeyValuePairs are &lt;Column, Direction&gt;.  Direction is True: Descending; False: Ascending</remarks>
        public string OrderBy(string Query, KeyValuePair<string, bool> Column1, params KeyValuePair<string, bool>[] Columns)
        {
            string res = Query + " ORDER BY " + Column1.Key + ((Column1.Value) ? " DESC" : " ASC");
            for (int i = 0; i < Columns.Length; i++)
                res += ", " + Columns[i].Key + ((Columns[i].Value) ? "` DESC" : " ASC");

            return res;
        }

        /// <summary>
        /// Makes a GROUP BY clause
        /// </summary>
        /// <param name="Query">The query to group</param>
        /// <param name="Column1">The first column to group by</param>
        /// <param name="Columns">Additional columns to group by</param>
        /// <returns>The Query</returns>
        public string GroupBy(string Query, string Column1, params string[] Columns)
        {
            return Query + " GROUP BY " + ColumnString(Column1, Columns);
        }

        /// <summary>
        /// Makes an AVG operation
        /// </summary>
        /// <param name="Column">The affected column</param>
        /// <returns>The Query</returns>
        public string Avg(string Column)
        {
            return "AVG(" + Column + ") ";
        }

        /// <summary>
        /// Makes a COUNT operation
        /// </summary>
        /// <param name="Column">The affected column</param>
        /// <returns>The Query</returns>
        public string Count(string Column)
        {
            return "COUNT(" + Column + ") ";
        }

        /// <summary>
        /// Makes a COUNT DISTINCT operation
        /// </summary>
        /// <param name="Column">The affected column</param>
        /// <returns>The Query</returns>
        public string CountDistinct(string Column)
        {
            return "COUNT(DISTINCT" + Column + ") ";
        }

        /// <summary>
        /// Makes a SUM operation
        /// </summary>
        /// <param name="Column">The affected column</param>
        /// <returns>The Query</returns>
        public string Sum(string Column)
        {
            return "SUM(" + Column + ") ";
        }

        /// <summary>
        /// Makes a MAX operation
        /// </summary>
        /// <param name="Column">The affected column</param>
        /// <returns>The Query</returns>
        public string Max(string Column)
        {
            return "MAX(" + Column + ") ";
        }

        /// <summary>
        /// Makes a MIN operation
        /// </summary>
        /// <param name="Column">The affected column</param>
        /// <returns>The Query</returns>
        public string Min(string Column)
        {
            return "MIN(" + Column + ") ";
        }

        /// <summary>
        /// Provides an ALIAS for a query
        /// </summary>
        /// <param name="Query">The query to alias</param>
        /// <param name="Alias">The alias to use</param>
        /// <returns>The Query</returns>
        public string Alias(string Query, string Alias)
        {
            return Query + " AS " + Alias;
        }

        /// <summary>
        /// Joins the results of two queries
        /// </summary>
        /// <param name="Query1">The query to alias</param>
        /// <param name="Query2">The alias to use</param>
        /// <param name="Condition">The join condition</param>
        /// <param name="JoinType">The type of join</param>
        /// <returns>The Query</returns>
        public string Join(string Query1, string Query2, string Condition, string JoinType)
        {
            return Query1 + " " + JoinType + " JOIN " + Query2;
        }

        /// <summary>
        /// Concatenates two or more fields
        /// </summary>
        /// <param name="Field1">The first field</param>
        /// <param name="Fields">Additional fields</param>
        /// <returns>The Query</returns>
        public string Concat(string Field1, params string[] Fields)
        {
            return "CONCAT(" + ColumnString(Field1, Fields) + ") ";
        }

        #region IPlugin Members

        public string Author
        {
            get { throw new Exception("The method or operation is not implemented."); }
        }

        public string Version
        {
            get { throw new Exception("The method or operation is not implemented."); }
        }

        public string Label
        {
            get { throw new Exception("The method or operation is not implemented."); }
        }

        #endregion
    }
}
