using System;
using System.Collections.Generic;
using System.Text;
using DuffBeer.Sensors.Interfaces;
using MrHomer.Plugins;

namespace DuffBeer.Sensors
{
    [MrHomerPlugin("Name", "Composite Data Filter")]
    [MrHomerPlugin("Description", "Filters the data with the first filter and feeds its output into the second filter.")]
    public class CompositeFilter : IDataFilter
    {
        private IDataFilter filter1;
        private IDataFilter filter2;
        #region IDataFilter Members

        public CompositeFilter(IDataFilter first, IDataFilter second)
        {
            if (!filter1.OutputType.Equals(filter2.InputType))
                throw new Exception("First filter's output type must match the second filter's input type.");
            filter1 = first;
            filter2 = second;
        }

        internal CompositeFilter()
        { }

        public object Filter(object data)
        {
            return filter2.Filter(filter1.Filter(data));
        }

        public byte[] SerializeConfiguration()
        {
            List<byte> bytes = new List<byte>();
            byte[] first = filter1.SerializeConfiguration();
            byte[] second = filter2.SerializeConfiguration();
            byte[] firsttype = new UTF8Encoding().GetBytes(filter1.GetType().FullName);
            byte[] secondtype = new UTF8Encoding().GetBytes(filter2.GetType().FullName);
            bytes.AddRange(BitConverter.GetBytes((Int32)firsttype.Length));
            bytes.AddRange(firsttype);
            bytes.AddRange(BitConverter.GetBytes((Int32)secondtype.Length));
            bytes.AddRange(secondtype);
            bytes.AddRange(BitConverter.GetBytes((Int32)first.Length));
            bytes.AddRange(first);
            bytes.AddRange(BitConverter.GetBytes((Int32)second.Length));
            bytes.AddRange(second);
            return bytes.ToArray();
        }

        public void DeserializeConfiguration(byte[] data)
        {
            byte[] first;
            byte[] second;
            byte[] firsttype;
            byte[] secondtype;

            int count = 0;

            firsttype = new byte[BitConverter.ToInt32(data, count)];
            count += sizeof(Int32);
            data.CopyTo(firsttype, count);
            count += firsttype.Length;

            secondtype = new byte[BitConverter.ToInt32(data, count)];
            count += sizeof(Int32);
            data.CopyTo(secondtype, count);
            count += secondtype.Length;

            first = new byte[BitConverter.ToInt32(data, count)];
            count += sizeof(Int32);
            data.CopyTo(first, count);
            count += first.Length;

            second = new byte[BitConverter.ToInt32(data, count)];
            count += sizeof(Int32);
            data.CopyTo(second, count);
            count += second.Length;

            PluginManager pm = PluginManager.getInstance();

            try
            {
                filter1 = pm.CreateInstance<IDataFilter>(Type.GetType(new UTF8Encoding().GetString(firsttype)), new object[] { });
            }
            catch (Exception e)
            {
                throw new Exception("Filter 1 Instantiation failed: " + e.Message);
            }
            try
            {
                filter2 = pm.CreateInstance<IDataFilter>(Type.GetType(new UTF8Encoding().GetString(secondtype)), new object[] { });
            }
            catch (Exception e)
            {
                throw new Exception("Filter 2 Instantiation failed: " + e.Message);
            }

            try
            {
                filter1.DeserializeConfiguration(first);
            }
            catch (Exception e)
            {
                throw new Exception("Filter 1 Serialization failed: " + e.Message);
            }

            try
            {
                filter2.DeserializeConfiguration(second);
            }
            catch (Exception e)
            {
                throw new Exception("Filter 2 Serialization failed: " + e.Message);
            }
        }

        public Type InputType
        {
            get { return filter1.InputType; }
        }

        public Type OutputType
        {
            get { return filter2.OutputType; }
        }

        #endregion

        #region IPlugin Members

        public string Author
        {
            get { return "Chris Scalabrini"; }
        }

        public string Version
        {
            get { return "1.0"; }
        }

        public string Label
        {
            get { return "Two-part Filter that feeds the results of the first into the second."; }
        }

        #endregion

        private int dbid;
        public int DBID
        {
            get
            {
                return dbid;
            }
            set
            {
                dbid = value;
            }

        }

        #region IDataFilter Members

        private Sensor sensor;
        public Sensor associatedSensor
        {
            get
            {
                return sensor;
            }
            set
            {
                sensor = value;
            }
        }

        #endregion
    }
}
