using System;
using System.Collections.Generic;
using System.Text;
using System.Threading;
using DuffBeer.Sensors.Interfaces;
using DuffBeer.Sensors.Utility;
using DuffBeer.Database.StructureTypes;
using DuffBeer.Database;
using MrHomer;

namespace DuffBeer.Sensors
{
    /// <summary>
    /// An actual sensor.  Only create this object when you are running the program.
    /// </summary>
    public class Sensor
    {
        #region Private Variables

        internal ISensor sensor;
        internal Dictionary<string, IDataFilter> filters;
        ThreadStart start;
        Thread thread;
        int sleeptime;

        public int refresh
        {
            get{return sleeptime;}
        }

        Boolean go;
        private string name;

        #endregion

        public string TypeName
        {
            get { return name; }
            set { name = value; }
        }

        #region Events

        /// <summary>
        /// Used by all events that need the Sensor.
        /// </summary>
        /// <param name="sensor">The sensor whose event being subscribed to.</param>
        public delegate void SensorDelegate(Sensor sensor);

        /// <summary>
        /// Sensor error.
        /// </summary>
        /// <param name="sensor">The sensor that threw an error.</param>
        /// <param name="e">The exception.</param>
        public delegate void SensorErrorDelegate(Sensor sensor, Exception e);

        /// <summary>
        /// Filter Error.
        /// </summary>
        /// <param name="f">The filter.</param>
        /// <param name="e">The exception.</param>
        public delegate void FilterErrorDelegate(IDataFilter f, Exception e);

        /// <summary>
        /// Upon a sensor refresh, this event is fired.
        /// </summary>
        public event SensorDelegate OnRefresh;

        /// <summary>
        /// Upon a sensor error, this event is fired.  In addition, the sensor reading is halted.
        /// </summary>
        public event SensorErrorDelegate SensorError;

        /// <summary>
        /// Upon a filter error, this event is fired.  The filter is also removed.  However, sensor reading
        /// will continue.
        /// </summary>
        public event FilterErrorDelegate FilterError;
        
        #endregion

        #region Constructors

        /// <summary>
        /// Creates an instance of the Sensor class.
        /// </summary>
        /// <param name="Sensor">The ISensor plugin object this sensor is going to use.</param>
        /// <param name="refresh">The sensor's refresh rate, in milliseconds.</param>
        public Sensor(ISensor Sensor, int refresh)
        {
            sensor = Sensor;
            filters = new Dictionary<string,IDataFilter>();
            sleeptime = refresh;
            start = new ThreadStart(Go);
        }

        /// <summary>
        /// Creates an instance of the Sensor class.
        /// </summary>
        /// <param name="Sensor">The ISensor plugin object this sensor is going to use.</param>
        /// <param name="refresh">The sensor's refresh rate, in milliseconds.</param>
        /// <param name="Filters">The set of filters to be used.</param>
        public Sensor(ISensor Sensor, int refresh, IEnumerable<IDataFilter> Filters)
        {
            sensor = Sensor;
            filters = new Dictionary<string,IDataFilter>();
            sleeptime = refresh;
            start = new ThreadStart(Go);
        }

        /// <summary>
        /// Creates an instance of the Sensor class.
        /// </summary>
        /// <param name="Sensor">The ISensor plugin object this sensor is going to use.</param>
        /// <param name="refresh">The sensor's refresh rate, in milliseconds.</param>
        /// <param name="config">The configuration for the sensor.</param>
        public Sensor(ISensor Sensor, int refresh, byte[] config)
        {
            sensor = Sensor;
            sensor.DeserializeConfiguration(config);
            filters = new Dictionary<string,IDataFilter>();
            sleeptime = refresh;
            start = new ThreadStart(Go);
        }

        /// <summary>
        /// Creates an instance of the Sensor class.
        /// </summary>
        /// <param name="Sensor">The ISensor plugin object this sensor is going to use.</param>
        /// <param name="refresh">The sensor's refresh rate, in milliseconds.</param>
        /// <param name="config">The configuration for the sensor.</param>
        /// <param name="Filters">The set of filters to be used.</param>
        public Sensor(ISensor Sensor, int refresh, byte[] config, IEnumerable<IDataFilter> Filters)
        {
            sensor = Sensor;
            sensor.DeserializeConfiguration(config);
            filters = new Dictionary<string,IDataFilter>();
            sleeptime = refresh;
            start = new ThreadStart(Go);
        }
        #endregion

        #region Public Methods

        /// <summary>
        /// Removes a filter from a sensor.
        /// </summary>
        /// <param name="name">The name of the filter to remove.</param>
        public void RemoveFilter(string name)
        {
            lock (filters)
            {
                if (filters.ContainsKey(name))
                {
                    filters[name].associatedSensor = null;
                    filters.Remove(name);
                }
            }
        }
        
        /// <summary>
        /// Adds a filter.
        /// </summary>
        /// <param name="filter">The filter to be used.</param>
        public void AddFilter(string name, IDataFilter filter)
        {
            if (filter == null)
                return;
            lock (filters)
            {
                if (!filters.ContainsKey(name))
                {
                    filters.Add(name, filter);
                    filter.associatedSensor = this;
                }
            }
        }

        /// <summary>
        /// Determines whether or not the named filter is on the sensor.
        /// </summary>
        /// <param name="name">The name of the filter.</param>
        /// <returns>True if the sensor contains the filter, false otherwise.</returns>
        public bool ContainsFilter(string name)
        {
            return filters.ContainsKey(name);
        }

        /// <summary>
        /// Returns the types of DataFilters in use by the sensor.
        /// </summary>
        /// <returns>A list of all filter types.</returns>
        public Dictionary<string,Type> GetFilterTypes()
        {
            Dictionary<string,Type> returnme = new Dictionary<string,Type>();
            lock (filters)
            {
                foreach (KeyValuePair<string, IDataFilter> f in filters)
                {
                    returnme.Add(f.Key,f.Value.GetType());
                }
            }
            return returnme;
        }

        /// <summary>
        /// Removes all filters.
        /// </summary>
        public void ClearFilters()
        {
            lock (filters)
            {
                foreach (IDataFilter f in filters.Values)
                {
                    f.associatedSensor = null;
                    FiltersTableEntry.Find("ID", f.DBID.ToString()).Delete();
                }
                filters.Clear();
            }
        }

        /// <summary>
        /// Returns the byte stream configurations of all filters, along with their type names.
        /// </summary>
        /// <returns>A Dictionary<string, byte[]> of filter type names and byte stream configurations.</returns>
        public Dictionary<string, KeyValuePair<string,byte[]>> FilterData()
        {
            Dictionary<string, KeyValuePair<string, byte[]>> returnme = new Dictionary<string, KeyValuePair<string, byte[]>>();
            lock (filters)
            {
                foreach (KeyValuePair<string, IDataFilter> df in filters)
                {
                    returnme.Add(df.Key, new KeyValuePair<string, byte[]>(df.Value.GetType().FullName, df.Value.SerializeConfiguration()));
                }
            }
            return returnme;
        }

        /// <summary>
        /// Returns the sensor's type name and its byte stream configuration.
        /// </summary>
        /// <returns>A KeyValuePair<string, byte[]> containing the sensor's type name and byte stream configuration.</returns>
        public KeyValuePair<string, byte[]> SensorData()
        {
            KeyValuePair<string, byte[]> returnme;
            lock (sensor)
            {
                returnme =  new KeyValuePair<string, byte[]>(sensor.GetType().FullName, sensor.SerializeConfiguration());
            }
            return returnme;
        }

        /// <summary>
        /// Starts the sensor readings.  If it's already going, will do nothing.
        /// </summary>
        public void Start()
        {
            if (!go)
            {
                go = true;
                thread = new Thread(start);
                thread.Start();
            }
        }

        /// <summary>
        /// Completes the next reading, and stops.  If it's already stopped, will do nothing.
        /// </summary>
        public void Stop()
        {
            if (go)
            {
                go = false;
            }
        }

        /// <summary>
        /// Set the sensor's refresh rate.  Can be set while the sensor is going.
        /// </summary>
        /// <param name="i">Sensor's new refresh rate, in milliseconds.</param>
        public void SetRefresh(int i)
        {
            sleeptime = i;
        }

        #endregion

        #region Private thread stuff

        /// <summary>
        /// The refresh thread.  Go!
        /// </summary>
        private void Go()
        {
            Thread.Sleep(sleeptime);

                lock (sensor)
                {
                    try
                    {
                        Filter(sensor.RequestData());
                    }
                    catch (Exception e)
                    {
                        Monitor.Exit(sensor);
                        if (SensorError != null)
                            SensorError(this, e);
                        go = false;
                    }
                }


            if (!go || Environment.HasShutdownStarted)
                return;
            else
                if(OnRefresh != null)
                    OnRefresh(this);

            lock (thread)
            {
                thread = new Thread(start);
                thread.Name = "Sensor " + name + " Updater";
                thread.Start();
            }
        }

        /// <summary>
        /// Filters the data.  Go!
        /// </summary>
        /// <param name="d">The data reading to filter.</param>
        private void Filter(object d)
        {
            Dictionary<string, Exception> errors = new Dictionary<string, Exception>();
            lock (filters)
            {
                foreach (KeyValuePair<string, IDataFilter> df in filters)
                {
                    try
                    {
                        df.Value.Filter(d);
                    }
                    catch (Exception e)
                    {
                        ErrorManager.getInstance().Catch(e);
                        errors.Add(df.Key, e);
                    }
                }
                foreach (KeyValuePair<string, Exception> k in errors)
                {
                    if(FilterError != null)
                        FilterError(filters[k.Key], k.Value);
                    filters.Remove(k.Key);
                }
            }

        }

        #endregion
    }
}
