using System;
using System.Collections.Generic;
using System.Text;
using System.Reflection;
using MrHomer.Plugins;
using MrHomer.Database;
using MrHomer.Database.Query;
using DuffBeer.Sensors.Interfaces;
using DuffBeer.Database.StructureTypes;
using DuffBeer.Database;
using MrHomer;

namespace DuffBeer.Sensors
{
    /// <summary>
    /// The sensor manager.  This is essentially an application of the PluginManager.
    /// </summary>
    public class SensorManager
    {
        #region Singleton

        static private SensorManager instance;
        /// <summary>
        /// Gets the SensorManager of the day.
        /// </summary>
        /// <returns>The sensor manager to end all sensor managers.</returns>
        static public SensorManager getInstance()
        {
            if (instance == null)
                instance = new SensorManager();
            return instance;
        }

        #endregion

        #region Private Variables

        private Dictionary<string, KeyValuePair<SensorInformation, Type>> sensors;
        private Dictionary<Type, string> sensorsBack;
        private Dictionary<string, KeyValuePair<FilterInformation, Type>> filters;
        private Dictionary<Type, string> filtersBack;
        private Dictionary<string, Sensor> activeSensors;
        private PluginManager pm;
        #endregion

        /// <summary>
        /// INVISIBLE! lulz.
        /// </summary>
        private SensorManager()
        {
            sensors = new Dictionary<string, KeyValuePair<SensorInformation, Type>>();
            sensorsBack = new Dictionary<Type, string>();
            filters = new Dictionary<string, KeyValuePair<FilterInformation, Type>>();
            filtersBack = new Dictionary<Type, string>();
            activeSensors = new Dictionary<string, Sensor>();
            pm = PluginManager.getInstance();
        }

        #region Public Methods

        /// <summary>
        /// Will load valid sensors and filters based on what the Plugin Manager has hanging around.
        /// </summary>
        public void LoadSensorsAndFilters()
        {
            Type isensor = typeof(int);
            Type idatafilter = typeof(int);
            
            foreach (Type t in pm.getInterfaces())
            {
                if (t.Name.ToLower() == "isensor")
                {
                    isensor = t;
                }
                if (t.Name.ToLower() == "idatafilter")
                {
                    idatafilter = t;
                }
            }

            if (isensor.Equals(typeof(int)) || idatafilter.Equals(typeof(int)))
            {
                throw new SensorException("Default Plugin Library was not loaded.");
            }

            foreach (Type t in pm.getPlugins(isensor))
            {
                try
                {
                    SensorInformation info = new SensorInformation(t);
                    sensors.Add(info.Name, new KeyValuePair<SensorInformation, Type>(info, t));
                    sensorsBack.Add(t, info.Name);
                }
                catch (SensorException e)
                {
                    ErrorManager.getInstance().Catch(e);
                }
                catch
                {  }
            }
            foreach (Type t in pm.getPlugins(idatafilter))
            {
                try
                {
                    FilterInformation info = new FilterInformation(t);
                    filters.Add(info.Name, new KeyValuePair<FilterInformation, Type>(info, t));
                    filtersBack.Add(t, info.Name);
                }
                catch (SensorException e)
                {
                    ErrorManager.getInstance().Catch(e);
                }
                catch
                { }
            }
        }

        /// <summary>
        /// Returns a list of all valid sensors and their information.
        /// </summary>
        /// <returns>A list of sensorinformation objects.</returns>
        public List<SensorInformation> GetSensorTypeInformation()
        {
            List<SensorInformation> returnme = new List<SensorInformation>();
            foreach (KeyValuePair<SensorInformation, Type> k in sensors.Values)
            {
                returnme.Add(k.Key);
            }
            return returnme;
        }

        /// <summary>
        /// Returns a list of all valid filters and their information.
        /// </summary>
        /// <returns>A list of filterinformation objects.</returns>
        public List<FilterInformation> GetFilterTypeInformation()
        {
            List<FilterInformation> returnme = new List<FilterInformation>();
            foreach (KeyValuePair<FilterInformation, Type> k in filters.Values)
            {
                returnme.Add(k.Key);
            }
            return returnme;
        }

        /// <summary>
        /// Will instantiate a new sensor based on a SensorInformation.Name, a label, and a refresh rate.
        /// </summary>
        /// <param name="Label">Your name for the sensor.</param>
        /// <param name="SensorTypeName">The type's SensorInformation.Name.</param>
        /// <param name="Refresh">The refresh rate.</param>
        /// <returns>True if successful, false otherwise.</returns>
        public bool InstantiateSensor(string Label, string Type, string SensorTypeName, int Refresh)
        {
            try
            {
                if (sensors.ContainsKey(SensorTypeName) && !activeSensors.ContainsKey(Label))
                {
                    Type t = sensors[SensorTypeName].Value;
                    Sensor s = new Sensor(pm.CreateInstance<ISensor>(t, null),
                                                                   Refresh);
                    activeSensors.Add(Label, s);
                    
                    s.sensor.SetName(Label);
                    s.TypeName = SensorTypeName;
                    SensorsTableEntry lol = new SensorsTableEntry(Label, Type, Refresh, s.sensor.GetUnits(), t.FullName, new byte[] { });
                    lol.Save();
                    return true;
                }
            }
            catch (Exception e)
            {
                ErrorManager.getInstance().Catch(e);
            }
            return false;
        }

        /// <summary>
        /// Adds a filter to a given sensor instantiation.
        /// </summary>
        /// <param name="Label">The sensor to which the filter is being added.</param>
        /// <param name="FilterTypeName">The FilterInformation.Name of the filter.</param>
        /// <returns>True if it's successful, false otherwise.</returns>
        public bool AddFilter(string Label, string FilterName, string FilterTypeName)
        {
            try
            {
                if (activeSensors.ContainsKey(Label) && filters.ContainsKey(FilterTypeName))
                {
                    SensorsTableEntry table = SensorsTableEntry.Find(SensorsTableEntry.NameColumn, Label);
                    if(table == null)
                        return false;

                    Type t = filters[FilterTypeName].Value;
                    activeSensors[Label].AddFilter(FilterName, pm.CreateInstance<IDataFilter>(t, null));

                    FiltersTableEntry f = new FiltersTableEntry(table.ID, t.FullName, new byte[] { });
                    f.Save();
                    return true;
                }
            }
            catch (Exception e)
            {
                ErrorManager.getInstance().Catch(e);
            }
            return false;
        }

        /// <summary>
        /// Removes a filter from the given sensor.
        /// </summary>
        /// <param name="Sensor">The sensor from which the filter is to be removed.</param>
        /// <param name="Filter">The filter to remove from said sensor.</param>
        /// <returns>True if the sensor exists, false otherwise.</returns>
        public bool RemoveFilter(string Sensor, string Filter)
        {
            if (activeSensors.ContainsKey(Sensor))
            {
                if(activeSensors[Sensor].ContainsFilter(Filter))
                {
                    FiltersTableEntry f = FiltersTableEntry.Find("ID", activeSensors[Sensor].filters[Filter].DBID.ToString());               
                }
                activeSensors[Sensor].RemoveFilter(Filter);
                return true;
            }
            return false;
        }

        /// <summary>
        /// Clears all the filters off a sensor.
        /// </summary>
        /// <param name="Label">The sensor to wipe.</param>
        /// <returns>True if successful, false otherwise.</returns>
        public bool ClearFilters(string Label)
        {
            if (activeSensors.ContainsKey(Label))
            {
                activeSensors[Label].ClearFilters();
                return true;
            }
            return false;
        }

        /// <summary>
        /// Returns all the Filters on a given sensor.
        /// </summary>
        /// <param name="Label">The sensor in question.</param>
        /// <returns>All the FilterInformation structores of filters on the sensor.</returns>
        public Dictionary<string,FilterInformation> GetFiltersOnSensor(string Label)
        {
            Dictionary<string,FilterInformation> returnme = new Dictionary<string,FilterInformation>();
            if (activeSensors.ContainsKey(Label))
            {
                Dictionary<string,Type> list = activeSensors[Label].GetFilterTypes();
                foreach (KeyValuePair<string, Type> t in list)
                {
                    if (filtersBack.ContainsKey(t.Value))
                    {
                        returnme.Add(t.Key, filters[filtersBack[t.Value]].Key);
                    }
                }
            }
            return returnme;
        }

        /// <summary>
        /// Gets all the sensors that have been created.
        /// </summary>
        /// <returns>A list of all the sensor labels.</returns>
        public List<string> GetCreatedSensors()
        {
            return new List<string>(activeSensors.Keys);
        }

        /// <summary>
        /// Gets the information for a specific sensor.
        /// </summary>
        /// <param name="sensor">The sensor of interest.</param>
        /// <returns>Information about the requested sensor</returns>
        public SensorInformation GetSensorInformation(string sensor)
        {
            if (activeSensors.ContainsKey(sensor))
            {
                return sensors[activeSensors[sensor].TypeName].Key;
            }
            else
                return null;
        }

        /// <summary>
        /// Starts a sensor.  If it is already started, will return true.
        /// </summary>
        /// <param name="Label">The sensor's label.</param>
        /// <returns>True if successful, false otherwise.</returns>
        public bool StartSensor(string Label)
        {
            if (activeSensors.ContainsKey(Label))
            {
                activeSensors[Label].Start();
                return true;
            }
            return false;
        }

        /// <summary>
        /// Stops a sensor.  If it is already stopped, will return true.
        /// </summary>
        /// <param name="Label">The sensor's label.</param>
        /// <returns>True if successful, false otherwise.</returns>
        public bool StopSensor(String Label)
        {
            if (activeSensors.ContainsKey(Label))
            {
                activeSensors[Label].Stop();
                return true;
            }
            return false;
        }

        /// <summary>
        /// Starts all sensors.  If some are already started, it won't affect them.
        /// </summary>
        public void StartAllSensors()
        {
            foreach (Sensor s in activeSensors.Values)
            {
                s.Start();
            }
        }

        /// <summary>
        /// Stops all sensors.  If some are already stopped, it won't affect them.
        /// </summary>
        public void StopAllSensors()
        {
            foreach (Sensor s in activeSensors.Values)
            {
                s.Stop();
            }
        }

        /// <summary>
        /// Sets the refresh rate of a sensor.
        /// </summary>
        /// <param name="Label">The sensor's label.</param>
        /// <param name="Refresh">The refresh rate.</param>
        /// <returns>True if successful, false otherwise.</returns>
        public bool SetRefresh(string Label, int Refresh)
        {
            if (activeSensors.ContainsKey(Label))
            {
                SensorsTableEntry table = SensorsTableEntry.Find(SensorsTableEntry.NameColumn, Label);
                table.RefreshRate = Refresh;
                table.Save();
                activeSensors[Label].SetRefresh(Refresh);
                return true;
            }
            return false;
        }


        public void WriteToDatabase()
        {
            foreach (KeyValuePair<string,Sensor> ss in activeSensors)
            {
                lock (ss.Value.filters)
                {
                    foreach (IDataFilter f in ss.Value.filters.Values)
                    {
                        FiltersTableEntry ts = FiltersTableEntry.Find(FiltersTableEntry.SensorColumn, ss.Key);
                        if (ts != null)
                        {
                            ts.State = f.SerializeConfiguration();
                            ts.Save();
                        }
                    }
                }
                SensorsTableEntry ste = SensorsTableEntry.Find(SensorsTableEntry.NameColumn, ss.Key);
                if (ste != null)
                {
                    ste.State = ss.Value.sensor.SerializeConfiguration();
                    ste.Save();
                }
            }
        }


        public void LoadFromDatabase()
        {
            IQueryBuilder q = DatabaseManager.getInstance().GetQueryBuilder(DuffBeerDatabaseConfig.DatabaseName);
            DataTable d = new DataTable(DatabaseManager.getInstance().FindDatabase(DuffBeerDatabaseConfig.DatabaseName), q, DuffBeerDatabaseConfig.SensorsTable, "ID");
            d.MoveFirst();
            do
            {
                
            }
            while (d.MoveNext());
        }

        #endregion

        #region Private Methods

        #endregion



    }
}
