using System;
using System.Collections.Generic;
using System.Text;
using System.Collections;

namespace MrHomer.Database
{
    public class DBResults
    {
        private List<string> indexMap;
        private List<Dictionary<string, string>> results;

        public DBResults()
        {
            indexMap = new List<string>();
            results = new List<Dictionary<string, string>>();
        }

        /// <summary>
        /// Creates a new database query result table from the given keys and data
        /// </summary>
        /// <param name="keys">The column names</param>
        /// <param name="values">The data in the table (as [row][col])</param>
        public DBResults(string[] keys, string[][] values)
        {
            if (values.Length > 0 && keys.Length != values[0].Length)
                throw new ArgumentException("When creating a results table, there must be the same number of keys as columns.");

            indexMap = new List<string>(keys);
            results = new List<Dictionary<string, string>>();

            for (int i = 0; i < values.Length; i++)
            {
                Dictionary<string, string> d = new Dictionary<string, string>();
                for (int k = 0; k < values[0].Length; k++)
                    d.Add(keys[k], values[i][k]);
                results.Add(d);
            }
        }

        /// <summary>
        /// Gets or sets a row of results from the table
        /// </summary>
        /// <param name="row">The row number to get</param>
        /// <returns>The row requested</returns>
        public Dictionary<string, string> this[int row]
        {
            get { return (results.Count > 0) ? results[row] : null; }
            set { results[row] = value; }
        }

        /// <summary>
        /// Gets or sets a specific data point from a table
        /// </summary>
        /// <param name="row">The row number</param>
        /// <param name="column">The column number</param>
        /// <returns>The requested data</returns>
        public string this[int row, int column]
        {
            get { return results[row][indexMap[column]]; }
            set { results[row][indexMap[column]] = value; }
        }

        /// <summary>
        /// Gets or sets a specific data point from a table
        /// </summary>
        /// <param name="row">The row number</param>
        /// <param name="column">The column name</param>
        /// <returns>The requested data</returns>
        public string this[int row, string column]
        {
            get { return results[row][column]; }
            set { results[row][column] = value; }
        }

        /// <summary>
        /// Adds an empty row to the table
        /// </summary>
        public void AddRow()
        {
            Dictionary<string, string> d = new Dictionary<string, string>();
            for (int i = 0; i < indexMap.Count; i++)
                d.Add(indexMap[i], "");

            results.Add(d);
        }

        /// <summary>
        /// Adds a row of data to the table using the existing keys
        /// </summary>
        /// <param name="data">The data to add</param>
        public void AddRow(string[] data)
        {
            if (data.Length != indexMap.Count) throw new ArgumentException("Number of columns must be the same as the number of keys");

            Dictionary<string, string> d = new Dictionary<string, string>();
            for (int i = 0; i < indexMap.Count; i++)
                d.Add(indexMap[i], data[i]);

            results.Add(d);
        }

        /// <summary>
        /// Adds a row of data to the table using the provided keys
        /// </summary>
        /// <param name="keys">The column names</param>
        /// <param name="data">The data to add</param>
        public void AddRow(string[] keys, string[] data)
        {
            if (data.Length != keys.Length) throw new ArgumentException("Number of columns must be the same as the number of keys");

            Dictionary<string, string> d = new Dictionary<string, string>();
            for (int i = 0; i < keys.Length; i++)
                d.Add(keys[i], data[i]);

            results.Add(d);
        }

        /// <summary>
        /// Gets the keys for the data
        /// </summary>
        /// <returns>The column names</returns>
        public string[] GetKeys()
        {
            return indexMap.ToArray();
        }
        /// <summary>
        /// Gets the number of rows in the table.
        /// </summary>
        /// <returns></returns>
        public int Length()
        {
            return results.Count;
        }

        public IEnumerator GetEnumerator()
        {
            return results.GetEnumerator();
        }
    }
}
