using System;
using System.Reflection;
using System.Collections.Generic;
using System.Text;
using MrHomer.Database.Query;

namespace MrHomer.Database
{
    public class DatabaseManager
    {
        private Dictionary<string, KeyValuePair<string, IDatabase>> databases;
        private Dictionary<string, IQueryBuilder> queries;
        private DatabaseAFactory factory;

        #region Singleton

        static private DatabaseManager instance;

        static public DatabaseManager getInstance()
        {
            if (instance == null)
                instance = new DatabaseManager();
            return instance;
        }

        #endregion

        private DatabaseManager()
        {
            databases = new Dictionary<string,  KeyValuePair<string, IDatabase>>();
            queries = new Dictionary<string, IQueryBuilder>();
            Plugins.PluginManager pm = Plugins.PluginManager.getInstance();
            pm.LoadInterfaceTypes(Assembly.GetExecutingAssembly());
            factory = new DatabaseAFactory();
        }

        /*
        /// <summary>
        /// Registers a new factory type with the manager.
        /// </summary>
        /// <param name="name">The label you wish to assign to the factory.</param>
        /// <param name="type">An instance of the factory.</param>
        /// <param name="builder">An instance of the query builder.</param>
        public void RegisterType(string name, IDatabaseFactory type, IQueryBuilder builder)
        {
            factory.RegisterType(name, type);
            if (!queries.ContainsKey(name))
            {
                queries.Add(name, builder);
            }
        }
        */

        /// <summary>
        /// Registers plugin types.
        /// </summary>
        public void RegisterTypes()
        {
            Plugins.PluginManager pm = Plugins.PluginManager.getInstance();
            List<Type> idatabase = pm.getPlugins(typeof(IDatabase));
            List<Type> iquerybuilder = pm.getPlugins(typeof(IQueryBuilder));
            List<Type> idatabasefactory = pm.getPlugins(typeof(IDatabaseFactory));

            Dictionary<string, Type[]> pain = new Dictionary<string, Type[]>();
            foreach (Type t in idatabase)
            {
                if (pm.EnforceAttributes("Required", t))
                {
                    foreach (Plugins.MrHomerPluginAttribute a in pm.getAttributes(t))
                    {
                        if (a.Name == "Type")
                        {
                            if (!pain.ContainsKey(a.Value))
                            {
                                pain.Add(a.Value, new Type[3]);
                                pain[a.Value][0] = t;
                                pain[a.Value][1] = null;
                                pain[a.Value][2] = null;
                            }
                        }
                    }
                }
            }
            foreach (Type t in iquerybuilder)
            {
                if (pm.EnforceAttributes("Required", t))
                {
                    foreach (Plugins.MrHomerPluginAttribute a in pm.getAttributes(t))
                    {
                        if (a.Name == "Type")
                        {
                            if (pain.ContainsKey(a.Value))
                            {
                                pain[a.Value][1] = t;
                            }
                        }
                    }
                }
            }
            foreach (Type t in idatabasefactory)
            {
                if (pm.EnforceAttributes("Required", t))
                {
                    foreach (Plugins.MrHomerPluginAttribute a in pm.getAttributes(t))
                    {
                        if (a.Name == "Type")
                        {
                            if (pain.ContainsKey(a.Value))
                            {
                                pain[a.Value][2] = t;
                            }
                        }
                    }
                }
            }
            List<string> removeme = new List<string>();
            foreach (KeyValuePair<string, Type[]> t in pain)
            {
                if (t.Value[1] == null || t.Value[2] == null)
                    removeme.Add(t.Key);
            }
            foreach (string s in removeme)
            {
                pain.Remove(s);
            }

            foreach (KeyValuePair<string, Type[]> t in pain)
            {
                queries.Add(t.Key, pm.CreateInstance<IQueryBuilder>(t.Value[1], new object[] { }));
                factory.RegisterType(t.Key, pm.CreateInstance<IDatabaseFactory>(t.Value[2], new object[] { }));
            }
        }

        /// <summary>
        /// Gets the database types.
        /// </summary>
        /// <returns>A list of the table types.</returns>
        public List<string> GetDatabaseTypes()
        {
            return new List<string>(queries.Keys);

        }

        /// <summary>
        /// Gets the QueryBuilder for the given database.
        /// </summary>
        /// <param name="name">The Database Name.</param>
        /// <returns>The IQueryBuilder object.</returns>
        public IQueryBuilder GetQueryBuilder(string name)
        {
            if (!databases.ContainsKey(name) || !queries.ContainsKey(databases[name].Key))
                return null;
            return queries[databases[name].Key];
        }


        /// <summary>
        /// Gets a database that has been connected
        /// </summary>
        /// <param name="name">The database to retrieve</param>
        /// <returns>The database object</returns>
        public IDatabase FindDatabase(string name)
        {
            if (databases.ContainsKey(name))
                return databases[name].Value;
            else
                throw new Exception("No connection to the given database exists.");
        }

        /// <summary>
        /// Establishes a connection to a database.
        /// </summary>
        /// <param name="name">The identifier for the database.</param>
        /// <param name="type">The type, provided earlier.</param>
        /// <param name="source">The data source for the connection (usually hostname:port) </param>
        /// <param name="username">The username for the connection</param>
        /// <param name="pass">The password for the given user</param>
        /// <param name="defaultdatabase">The database to use></param>
        /// <param name="arguments">Any arguments that might be needed.  Usually null.</param>
        public void ConnectDatabase(string name, string type, string source, string username, string pass, string defaultdatabase, object[] arguments)
        {
            IDatabase d = factory.GetDatabase(type, arguments);
            if (d == null)
                return;
            d.Connect(source, username, pass, defaultdatabase);
            databases.Add(name, new KeyValuePair<string, IDatabase>(type, d));
        }

        /// <summary>
        /// Terminates a connection to the database
        /// </summary>
        /// <param name="which"></param>
        public void DisconnectDatabase(string which)
        {
            if (databases.ContainsKey(which))
            {
                databases[which].Value.Disconnect();
                databases.Remove(which);
            }
        }

        /// <summary>
        /// Executes a query against a given database
        /// </summary>
        /// <param name="which">The database to query</param>
        /// <param name="query">The query string to run</param>
        /// <returns>The result of the query</returns>
        /// <remarks>This operation returns <code>null</code> if the database is not connected</remarks>
        public DBResults QueryDatabase(string which, string query)
        {
            if (databases.ContainsKey(which) && databases[which].Value.IsConnected())
                return databases[which].Value.Query(query);
            else
                return null;
        }
    }
}
