using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.Text;
using System.Net;
using System.Net.Sockets;
using System.Collections;

namespace AsyncSockets
{
    /// <summary>
    /// Provides asynchronous access to sockets
    /// </summary>
    public class AsyncSocket
    {

        #region Local Types

        public struct State
        {
            public EndPoint EndPoint;
            public SocketFlags SocketFlags;
            public IPPacketInformation PacketInformation;
            internal byte[] Buffer;
        }

        #endregion

        #region Data Fields

        /// <summary>
        /// The underlying network socket
        /// </summary>
        private Socket mySocket;

        #endregion

        #region Constructors

        /// <summary>
        /// Creates a new asynchronous socket
        /// </summary>
        /// <param name="socketInformation">Information about this socket</param>
        public AsyncSocket(SocketInformation socketInformation)
        {
            mySocket = new Socket(socketInformation);
        }

        /// <summary>
        /// Creates a new asynchronous socket
        /// </summary>
        /// <param name="addressFamily">The address family in use by the socket</param>
        /// <param name="protocolType">The protocol in use by the socket</param>
        /// <param name="socketType">The type of this socket</param>
        public AsyncSocket(AddressFamily addressFamily, SocketType socketType, ProtocolType protocolType)
        {
            mySocket = new Socket(addressFamily, socketType, protocolType);
        }

        /// <summary>
        /// Converts a normal socket into an asynchronous socket
        /// </summary>
        /// <param name="socket">The existing socket</param>
        private AsyncSocket(Socket socket)
        {
            mySocket = socket;
        }

        #endregion

        #region Properties

        /// <summary>
        /// Provides access to the base socket
        /// </summary>
        internal Socket BaseSocket
        {
            get { return mySocket; }
        }

        /// <summary>
        /// Gets the address family of the Socket.
        /// </summary>
        public AddressFamily AddressFamily
        {
            get { return mySocket.AddressFamily; }
        }

        /// <summary>
        /// Gets the amount of data that has been received from the network and is available to be read.
        /// </summary>
        public int Available
        {
            get { return mySocket.Available; }
        }

        /// <summary>
        /// Gets or sets a Boolean value that specifies whether the Socket allows Internet Protocol (IP) datagrams to be fragmented.
        /// </summary>
        public bool DontFragment
        {
            get { return mySocket.DontFragment; }
            set { mySocket.DontFragment = value; }
        }

        /// <summary>
        /// Gets or sets a Boolean value that specifies whether the Socket can send or receive broadcast packets.
        /// </summary>
        public bool EnableBroadcast
        {
            get { return mySocket.EnableBroadcast; }
            set { mySocket.EnableBroadcast = true; }
        }

        /// <summary>
        /// Gets or sets a Boolean value that specifies whether the Socket allows only one process to bind to a port.
        /// </summary>
        public bool ExclusiveAddressUse
        {
            get { return mySocket.ExclusiveAddressUse; }
            set { mySocket.ExclusiveAddressUse = value; }
        }

        /// <summary>
        /// Gets the operating system handle for the Socket.
        /// </summary>
        public IntPtr Handle
        {
            get { return mySocket.Handle; }
        }

        /// <summary>
        /// Gets a value that indicates whether the Socket is bound to a specific local port.
        /// </summary>
        public bool IsBound
        {
            get { return mySocket.IsBound; }
        }

        /// <summary>
        /// Gets a value that indicates whether a Socket is connected to a remote host as of the last Send or Receive operation.
        /// </summary>
        public bool IsConnected
        {
            get { return mySocket.Connected; }
        }

        /// <summary>
        /// Gets or sets a value that specifies whether the Socket will delay closing a socket in an attempt to send all pending data.
        /// </summary>
        public LingerOption LingerState
        {
            get { return mySocket.LingerState; }
        }

        /// <summary>
        /// Gets the local endpoint.
        /// </summary>
        public EndPoint LocalEndPoint
        {
            get { return mySocket.LocalEndPoint; }
        }

        /// <summary>
        /// Gets or sets a value that specifies whether outgoing multicast packets are delivered to the sending application.
        /// </summary>
        public bool MulticastLoopback
        {
            get { return mySocket.MulticastLoopback; }
            set { mySocket.MulticastLoopback = value; }
        }

        /// <summary>
        /// Gets or sets a Boolean value that specifies whether the stream Socket is using the Nagle algorithm.
        /// </summary>
        public bool NoDelay
        {
            get { return mySocket.NoDelay; }
            set { mySocket.NoDelay = value; }
        }
        
        /// <summary>
        /// Sets or retrieves the protocol type of this socket
        /// </summary>
        public ProtocolType ProtocolType
        {
            get { return mySocket.ProtocolType; }
        }

        /// <summary>
        /// Gets or sets a value that specifies the size of the receive buffer of the Socket.
        /// </summary>
        public int ReceiveBufferSize
        {
            get { return mySocket.ReceiveBufferSize; }
            set { mySocket.ReceiveBufferSize = value; }
        }

        /// <summary>
        /// Gets or sets a value that specifies the amount of time after which a synchronous Receive call will time out.
        /// </summary>
        public int ReceiveTimeout
        {
            get { return mySocket.ReceiveTimeout; }
            set { mySocket.ReceiveTimeout = value; }
        }

        /// <summary>
        /// Gets the remote endpoint.
        /// </summary>
        public EndPoint RemoteEndPoint
        {
            get { return mySocket.RemoteEndPoint; }
        }

        /// <summary>
        /// Gets or sets a value that specifies the size of the send buffer of the Socket.
        /// </summary>
        public int SendBufferSize
        {
            get { return mySocket.SendBufferSize; }
            set { mySocket.SendBufferSize = value; }
        }

        /// <summary>
        /// Gets or sets a value that specifies the amount of time after which a synchronous Send call will time out.
        /// </summary>
        public int SendTimeout
        {
            get { return mySocket.SendTimeout; }
            set { mySocket.SendTimeout = value; }
        }

        /// <summary>
        /// Gets the type of the Socket.
        /// </summary>
        public SocketType SocketType
        {
            get { return mySocket.SocketType; }
        }

        /// <summary>
        /// Gets or sets a value that specifies the Time To Live (TTL) value of Internet Protocol (IP) packets sent by the Socket.
        /// </summary>
        public short Ttl
        {
            get { return mySocket.Ttl; }
            set { mySocket.Ttl = value; }
        }

        /// <summary>
        /// Specifies whether the socket should only use Overlapped I/O mode.
        /// </summary>
        public bool UseOnlyOverlappedIO
        {
            get { return mySocket.UseOnlyOverlappedIO; }
            set { mySocket.UseOnlyOverlappedIO = value; }
        }

        #endregion

        #region Events

        public delegate void AsyncSocketEvent(Object sender, SocketEventArgs e);
        public delegate void AsyncSocketConnectEvent(Object sender, SocketConnectEventArgs e);
        public delegate void AsyncSocketDataEvent(Object sender, SocketDataEventArgs e);
        public delegate void AsyncSocketErrorEvent(Object sender, AsyncSocketException e);

        /// <summary>
        /// Raised when a connection to a remote host is completed
        /// </summary>
        public event AsyncSocketEvent Connected;
        /// <summary>
        /// Raised when the connection to the remote host is terminated
        /// </summary>
        public event AsyncSocketEvent Disconnected;

        /// <summary>
        /// Raised when a connection to a listening socket is accepted
        /// </summary>
        public event AsyncSocketConnectEvent AcceptedConnection;

        /// <summary>
        /// Raised when a socket has completed sending data
        /// </summary>
        public event AsyncSocketDataEvent DataSent;
        /// <summary>
        /// Raised when a socket has completed receiving data
        /// </summary>
        public event AsyncSocketDataEvent DataReceived;

        /// <summary>
        /// Raised when an error occurs in the socket
        /// </summary>
        public event AsyncSocketErrorEvent SocketError;

        #endregion

        #region Functions

        /// <summary>
        /// Raises the SocketError event
        /// </summary>
        /// <param name="message">The text of the error</param>
        /// <param name="e">The source of the error</param>
        private void OnError(string message, Exception e)
        {
            if (SocketError != null)
                SocketError(this, new AsyncSocketException(message, e));
        }

        /// <summary>
        /// Decodes the data received from the socket and raises the DataReceived event
        /// </summary>
        /// <param name="bytes">The number of bytes of data received</param>
        /// <param name="state">The socket state at the time the receive was requested</param>
        private void OnReceive(int bytes, State state)
        {
            if (DataReceived != null) DataReceived(this, new SocketDataEventArgs(mySocket.LocalEndPoint, state.EndPoint, bytes, state.Buffer, state));
        }

        /// <summary>
        /// Raises the DataSent event
        /// </summary>
        /// <param name="bytes">The number of bytes sent</param>
        /// <param name="state">The socket state at the time the send was requested</param>
        private void OnSend(int bytes, State state)
        {
            if (DataSent != null) DataSent(this, new SocketDataEventArgs(mySocket.LocalEndPoint, state.EndPoint, bytes, null, state));
        }

        /// <summary>
        /// Binds the socket to a local endpoint
        /// </summary>
        /// <param name="localEP">The endpoint to which to bind</param>
        public void Bind(EndPoint localEP)
        {
            if (mySocket.IsBound)
                throw new AsyncSocketException("This socket is already bound to an address");

            try
            {
                mySocket.Bind(localEP);
            }
            catch (SocketException e)
            {
                throw new AsyncSocketException("Failed to bind to the given endpoint", e);
            }
        }

        /// <summary>
        /// Listens for new incoming connections
        /// </summary>
        /// <param name="backlog">Maximum number of incoming requests to queue</param>
        public void Listen(int backlog) {
            mySocket.Listen(backlog);
        }

        /// <summary>
        /// Closes the socket
        /// </summary>
        public void Close()
        {
            mySocket.Close();
        }

        /// <summary>
        /// Closes the socket
        /// </summary>
        /// <param name="timeout">The time to wait</param>
        public void Close(int timeout)
        {
            mySocket.Close(timeout);
        }

        /// <summary>
        /// Disables sends and receives on a Socket. 
        /// </summary>
        /// <param name="how">One of the SocketShutdown values that specifies the operation that will no longer be allowed.</param>
        public void Shutdown(SocketShutdown how)
        {
            mySocket.Shutdown(how);
        }

        #endregion

        #region Async Functions

        #region Connect / Disconnect

        /// <summary>
        /// Initiates a connection to a remote host
        /// </summary>
        /// <param name="remoteEP">The end point to which a connection should be established</param>
        /// <exception cref="AsyncSockets.AsyncSocketException">The connection could not be established</exception>
        public void Connect(EndPoint remoteEP)
        {
            try
            {
                mySocket.BeginConnect(remoteEP, new AsyncCallback(Connect_Completed), null);
            }
            catch (SocketException e)
            {
                throw new AsyncSocketException("Failed to establish a connection to the remote host.", e);
            }
        }

        /// <summary>
        /// Initiates a connection to a remote host
        /// </summary>
        /// <param name="address">The IP address of the remote host</param>
        /// <param name="port">The port number of the remote host</param>
        /// <exception cref="AsyncSockets.AsyncSocketException">The connection could not be established</exception>
        public void Connect(IPAddress address, int port)
        {
            try
            {
                mySocket.BeginConnect(address, port, new AsyncCallback(Connect_Completed), null);
            }
            catch (SocketException e)
            {
                throw new AsyncSocketException("Failed to establish a connection to the remote host.", e);
            }
        }

        /// <summary>
        /// Initiates a connection to a remote host
        /// </summary>
        /// <param name="addresses">The IP addresses of the remote host</param>
        /// <param name="port">The port number of the remote host</param>
        /// <exception cref="AsyncSockets.AsyncSocketException">The connection could not be established</exception>
        public void Connect(IPAddress[] addresses, int port)
        {
            try
            {
                mySocket.BeginConnect(addresses, port, new AsyncCallback(Connect_Completed), null);
            }
            catch (SocketException e)
            {
                throw new AsyncSocketException("Failed to establish a connection to the remote host.", e);
            }
        }

        /// <summary>
        /// Initiates a connection to a remote host
        /// </summary>
        /// <param name="host">The host name of the remote host</param>
        /// <param name="port">The port number of the remote host</param>
        /// <exception cref="AsyncSockets.AsyncSocketException">The connection could not be established</exception>
        public void Connect(string host, int port)
        {
            try
            {
                mySocket.BeginConnect(host, port, new AsyncCallback(Connect_Completed), null);
            }
            catch (SocketException e)
            {
                throw new AsyncSocketException("Failed to establish a connection to the remote host.", e);
            }
        }

        /// <summary>
        /// Closes the connection to a remote host
        /// </summary>
        /// <param name="reuseSocket">Indicates if this socket can be reused</param>
        /// <exception cref="AsyncSockets.AsyncSocketException">The connection could not be terminated</exception>
        public void Disconnect(bool reuseSocket)
        {
            try
            {
                mySocket.BeginDisconnect(reuseSocket, new AsyncCallback(Disconnect_Completed), null);
            }
            catch (SocketException e)
            {
                throw new AsyncSocketException("Failed to close the connection to the remote host.", e);
            }
        }

        #endregion

        #region Receive

        /// <summary>
        /// Receives data from the remote host
        /// </summary>
        /// <param name="size">The number of bytes to receive</param>
        /// <param name="socketFlags">Any flags describing the socket</param>
        /// <exception cref="AsyncSockets.AsyncSocketException">Attempting to retrieve data from a closed socket</exception>
        public void Receive(int size, SocketFlags socketFlags)
        {
            if (!mySocket.Connected)
                throw new AsyncSocketException("Failed to receive data: Socket is closed");

            try
            {
                State s = new State();
                s.EndPoint = mySocket.RemoteEndPoint;
                s.SocketFlags = socketFlags;
                s.Buffer = new byte[size];

                mySocket.BeginReceive(s.Buffer, 0, size, socketFlags, new AsyncCallback(Recv_Completed), s);
            }
            catch (SocketException e)
            {
                throw new AsyncSocketException("Failed to receive data from the remote host.", e);
            }
        }

        /// <summary>
        /// Receives data from a specified remote host
        /// </summary>
        /// <param name="size">The number of bytes to receive</param>
        /// <param name="socketFlags">Any flags describing the socket</param>
        /// <param name="remoteEP">The remote host</param>
        /// <exception cref="AsyncSocketException"></exception>
        public void ReceiveFrom(int size, SocketFlags socketFlags, ref EndPoint remoteEP)
        {
            try
            {
                State s = new State();
                s.EndPoint = remoteEP;
                s.SocketFlags = socketFlags;
                s.Buffer = new byte[size];

                mySocket.BeginReceiveFrom(s.Buffer, 0, size, socketFlags, ref remoteEP, new AsyncCallback(RecvFrom_Completed), s);
            }
            catch (SocketException e)
            {
                throw new AsyncSocketException("Failed to receive data from the remote host.", e);
            }
        }

        /// <summary>
        /// Receives a message from a specific remote host
        /// </summary>
        /// <param name="size">The number of bytes to receive</param>
        /// <param name="socketFlags">Any flags describing the socket</param>
        /// <param name="remoteEP">The remote host</param>
        /// <exception cref="AsyncSocketException"></exception>
        public void ReceiveMessageFrom(int size, SocketFlags socketFlags, ref EndPoint remoteEP)
        {
            try
            {
                State s = new State();
                s.EndPoint = remoteEP;
                s.SocketFlags = socketFlags;
                s.Buffer = new byte[size];

                mySocket.BeginReceiveMessageFrom(s.Buffer, 0, size, socketFlags, ref remoteEP, new AsyncCallback(RecvMsg_Completed), s);
            }
            catch (SocketException e)
            {
                throw new AsyncSocketException("Failed to receive data from the remote host.", e);
            }
        }

        #endregion

        #region Send

        /// <summary>
        /// Send data to the remote host
        /// </summary>
        /// <param name="buffer">The data to send</param>
        public void Send(byte[] buffer)
        {
            Send(buffer, 0, buffer.Length, SocketFlags.None);
        }

        /// <summary>
        /// Send data to the remote host
        /// </summary>
        /// <param name="buffer">The data to send</param>
        /// <param name="socketFlags">Any flags for the socket</param>
        public void Send(byte[] buffer, SocketFlags socketFlags)
        {
            Send(buffer, 0, buffer.Length, socketFlags);
        }

        /// <summary>
        /// Send data to the remote host
        /// </summary>
        /// <param name="buffer">The data to send</param>
        /// <param name="size">The size of the data to send</param>
        /// <param name="socketFlags">Any flags for the socket</param>
        public void Send(byte[] buffer, int size, SocketFlags socketFlags)
        {
            Send(buffer, 0, size, socketFlags);
        }

        /// <summary>
        /// Send data to the remote host
        /// </summary>
        /// <param name="buffer">The data to send</param>
        /// <param name="size">The size of the data to send</param>
        /// <param name="offset">The offset of the data within the buffer</param>
        /// <param name="socketFlags">Any flags for the socket</param>
        public void Send(byte[] buffer, int offset, int size, SocketFlags socketFlags)
        {
            if (!mySocket.Connected)
                throw new AsyncSocketException("Failed to send data: Socket is closed");

            try
            {
                State s = new State();
                s.EndPoint = mySocket.RemoteEndPoint;
                s.SocketFlags = socketFlags;
                s.Buffer = new byte[size];
                for (int i = 0; i < size; i++)
                    s.Buffer[i] = buffer[i + offset];

                mySocket.BeginSend(buffer, offset, size, socketFlags, new AsyncCallback(Send_Completed), s);
            }
            catch (SocketException e)
            {
                throw new AsyncSocketException("Failed to send data to the remote host.", e);
            }
        }

        /// <summary>
        /// Send data to a specific remote host
        /// </summary>
        /// <param name="buffer">The data to send</param>
        /// <param name="remoteEP">The remote host</param>
        public void SendTo(byte[] buffer, EndPoint remoteEP)
        {
            SendTo(buffer, 0, buffer.Length, SocketFlags.None, remoteEP);
        }

        /// <summary>
        /// Send data to the remote host
        /// </summary>
        /// <param name="buffer">The data to send</param>
        /// <param name="socketFlags">Any flags for the socket</param>
        /// <param name="remoteEP">The remote host</param>
        public void SendTo(byte[] buffer, SocketFlags socketFlags, EndPoint remoteEP)
        {
            SendTo(buffer, 0, buffer.Length, socketFlags, remoteEP);
        }

        /// <summary>
        /// Send data to the remote host
        /// </summary>
        /// <param name="buffer">The data to send</param>
        /// <param name="size">The size of the data to send</param>
        /// <param name="socketFlags">Any flags for the socket</param>
        /// <param name="remoteEP">The remote host</param>
        public void SendTo(byte[] buffer, int size, SocketFlags socketFlags, EndPoint remoteEP)
        {
            SendTo(buffer, 0, size, socketFlags, remoteEP);
        }

        /// <summary>
        /// Send data to the remote host
        /// </summary>
        /// <param name="buffer">The data to send</param>
        /// <param name="size">The size of the data to send</param>
        /// <param name="offset">The offset of the data within the buffer</param>
        /// <param name="socketFlags">Any flags for the socket</param>
        /// <param name="remoteEP">The remote host</param>
        public void SendTo(byte[] buffer, int offset, int size, SocketFlags socketFlags, EndPoint remoteEP)
        {
            try
            {
                State s = new State();
                s.EndPoint = remoteEP;
                s.SocketFlags = socketFlags;
                s.Buffer = new byte[size];
                for (int i = 0; i < size; i++)
                    s.Buffer[i] = buffer[i + offset];

                mySocket.BeginSendTo(buffer, offset, size, socketFlags, remoteEP, new AsyncCallback(SendTo_Completed), s);
            }
            catch (SocketException e)
            {
                throw new AsyncSocketException("Failed to send data to the remote host.", e);
            }
        }
        
        /// <summary>
        /// Sends a file to the remote host
        /// </summary>
        /// <param name="filename">The name of the file to send</param>
        public void SendFile(string filename)
        {
            if (!mySocket.Connected)
                throw new AsyncSocketException("Failed to send file: Socket is closed");

            try
            {
                State s = new State();
                s.EndPoint = mySocket.RemoteEndPoint;
                s.SocketFlags = SocketFlags.None;

                mySocket.BeginSendFile(filename, new AsyncCallback(Send_Completed), s);
            }
            catch (SocketException e)
            {
                throw new AsyncSocketException("Failed to send file to the remote host.", e);
            }
        }

        /// <summary>
        /// Sends a file and the data buffers to the remote host
        /// </summary>
        /// <param name="filename">A String that contains the path and name of the file to be sent.</param>
        /// <param name="preBuffer">A Byte array that contains data to be sent before the file is sent.</param>
        /// <param name="postBuffer">A Byte array that contains data to be sent after the file is sent.</param>
        /// <param name="flags">One or more of TransmitFileOptions values.</param>
        public void SendFile(string filename, byte[] preBuffer, byte[] postBuffer, TransmitFileOptions flags)
        {
            if (!mySocket.Connected)
                throw new AsyncSocketException("Failed to send file: Socket is closed");

            try
            {
                State s = new State();
                s.EndPoint = mySocket.RemoteEndPoint;
                s.SocketFlags = SocketFlags.None;

                mySocket.BeginSendFile(filename, preBuffer, postBuffer, flags, new AsyncCallback(Send_Completed), s);
            }
            catch (SocketException e)
            {
                throw new AsyncSocketException("Failed to send file to the remote host.", e);
            }
        }

        #endregion

        #region Host

        /// <summary>
        /// Accepts a pending connection
        /// </summary>
        public void Accept()
        {
            try
            {
                mySocket.BeginAccept(new AsyncCallback(Accept_Completed), null);
            }
            catch (SocketException e)
            {
                throw new AsyncSocketException("Failed to accept incoming connection", e);
            }
        }

        #endregion

        #endregion

        #region Async Callbacks

        /// <summary>
        /// Finalizes the Connect operation
        /// </summary>
        private void Connect_Completed(IAsyncResult res)
        {
            try
            {
                mySocket.EndConnect(res);
                if (Connected != null) Connected(this, new SocketEventArgs(mySocket.LocalEndPoint, mySocket.RemoteEndPoint));
            }
            catch (Exception e)
            {
                OnError("Failed to establish a connection to the remote host.", e);
            }
        }

        /// <summary>
        /// Finalizes the Disconnect operation
        /// </summary>
        private void Disconnect_Completed(IAsyncResult res)
        {
            try
            {
                mySocket.EndDisconnect(res);
                if(Disconnected != null) Disconnected(this, new SocketEventArgs(mySocket.LocalEndPoint, mySocket.RemoteEndPoint));
            }
            catch (Exception e)
            {
                OnError("Failed to close the connection to the remote host.", e);
            }
        }

        /// <summary>
        /// Finalizes the Receive operation
        /// </summary>
        private void Recv_Completed(IAsyncResult res)
        {
            try
            {
                int bytes = mySocket.EndReceive(res);

                if (bytes == 0)
                {
                    Shutdown(SocketShutdown.Both);
                    if (Disconnected != null) Disconnected(this, new SocketEventArgs(mySocket.LocalEndPoint, mySocket.RemoteEndPoint));
                }
                else
                    OnReceive(bytes, (State)res.AsyncState);
            }
            catch (Exception e)
            {
                OnError("Failed to receive data from remote host.", e);
            }
        }

        /// <summary>
        /// Finalizes the ReceiveFrom operation
        /// </summary>
        private void RecvFrom_Completed(IAsyncResult res)
        {
            try
            {
                State s = (State)res.AsyncState;
                int bytes = mySocket.EndReceiveFrom(res, ref s.EndPoint);

                OnReceive(bytes, s);
            }
            catch (Exception e)
            {
                OnError("Failed to receive data from remote host.", e);
            }
        }

        /// <summary>
        /// Finalizes the ReceiveMessageFrom operation
        /// </summary>
        private void RecvMsg_Completed(IAsyncResult res)
        {
            try
            {
                State s = (State)res.AsyncState;
                int bytes = mySocket.EndReceiveMessageFrom(res, ref s.SocketFlags, ref s.EndPoint, out s.PacketInformation);
            }
            catch (Exception e)
            {
                OnError("Failed to receive message from remote host.", e);
            }
        }

        /// <summary>
        /// Finalizes the Send operation
        /// </summary>
        private void Send_Completed(IAsyncResult res)
        {
            try
            {
                int bytes = mySocket.EndSend(res);

                OnSend(bytes, (State)res.AsyncState);
            }
            catch (Exception e)
            {
                OnError("Failed to transfer data to remote host.", e);
            }
        }

        /// <summary>
        /// Finalizes the SendFile operation
        /// </summary>
        private void SendFile_Completed(IAsyncResult res)
        {
            try
            {
                mySocket.EndSendFile(res);

                OnSend(0, (State)res.AsyncState);
            }
            catch (Exception e)
            {
                SocketError(this, new AsyncSocketException("Failed to send file to remote host.", e));
            }
        }

        /// <summary>
        /// Finalizes the SendTo operation
        /// </summary>
        private void SendTo_Completed(IAsyncResult res)
        {
            try
            {
                int bytes = mySocket.EndSendTo(res);

                OnSend(bytes, (State)res.AsyncState);
            }
            catch (Exception e)
            {
                OnError("Failed to send data to remote host.", e);
            }
        }

        /// <summary>
        /// Finalizes the Accept operation
        /// </summary>
        private void Accept_Completed(IAsyncResult res)
        {
            try
            {
                Socket m = mySocket.EndAccept(res);

                if (AcceptedConnection != null) AcceptedConnection(this, new SocketConnectEventArgs(new AsyncSocket(m)));
            }
            catch (Exception e)
            {
                OnError("Failed to accept an incoming connection", e);
            }
        }

        #endregion
    }
}
