using System;
using System.ComponentModel;
using System.Collections.Generic;
using System.Text;
using System.Threading;
using System.Windows;
using AsyncSockets;
using System.Net.Sockets;

namespace MrHomer.Networking
{
    /// <summary>
    /// A connection manager that leaves security up to you.
    /// </summary>
    public class ConnectionManager : IDisposable
    {
        #region Singleton
        static private ConnectionManager instance;

        /// <summary>
        /// Returns an instance of the ConnectionManager.
        /// </summary>
        /// <returns>The current ConnectionManager instance, if there is one.  Otherwise, creates one and returns it.</returns>
        static public ConnectionManager getInstance()
        {
            if (instance == null)
                instance = new ConnectionManager();
            return instance;
        }
        #endregion

        #region Variables

        /* *
         * All manner of connections are here, including active connections (With forward and reverse
         * lookup for the AsyncSocket class), Inactive Connections (associated with credentials), and
         * connections listening on various ports.
         * */

        //For connections that are validated
        private Dictionary<string, AsyncSocket> ActiveConnectionsReference;
        private Dictionary<AsyncSocket, string> ActiveConnectionsBackReference;
        private Dictionary<string, KeyValuePair<AsyncSocket, byte[]>> InactiveConnections;
        private List<string> DroppedConnections;
        private List<string> FailedConnections;

        //For requesting connections
        private Dictionary<AsyncSocket, int> ListeningForConnection;
        private Dictionary<AsyncSocket, string> AwaitingCredentials;

        //For making connections
        private Dictionary<AsyncSocket, string> WaitingForConnection;
        private Dictionary<AsyncSocket, string> PendingApproval;

        private Dictionary<string, List<byte[]>> DataReceived;

        //Threads to go away on exit
        private List<Thread> killers;
        private List<Thread> spammers;
        private List<Thread> droppers;

        /* *
         * Credentials and other standard packets.
         * */
        private byte[] credentials;
        private byte[] CredentialRequestPacket;

        /* *
         * Additional useful junk
         * */
        private bool dropontimeout;
        private uint dropouttimer;

        static private ulong log;


        #endregion

        public delegate void Connection(string s);
        public event Connection ConnectionActivated;

        /*
         * The Delegate and event for notifying the console of new connections 
         */
        public delegate void ConnectNotification(string s);
        public event ConnectNotification NotifyOfClientConnection;

        /*
        * The Delegate and event for notifying the console of disconnections 
        */
        public delegate void DisconnectNotification(string s);
        public event DisconnectNotification NotifyOfClientDisconnection;

        private ConnectionManager()
        {
            ActiveConnectionsReference = new Dictionary<string, AsyncSocket>();
            ActiveConnectionsBackReference = new Dictionary<AsyncSocket, string>();
            InactiveConnections = new Dictionary<string, KeyValuePair<AsyncSocket, byte[]>>();
            DroppedConnections = new List<string>();
            ListeningForConnection = new Dictionary<AsyncSocket, int>();
            AwaitingCredentials = new Dictionary<AsyncSocket, string>();
            WaitingForConnection = new Dictionary<AsyncSocket, string>();
            PendingApproval = new Dictionary<AsyncSocket, string>();
            DataReceived = new Dictionary<string, List<byte[]>>();
            dropontimeout = false;
            dropouttimer = 10000;
            ConnectionManager.log = 0;
            credentials = new byte[0];
            CredentialRequestPacket = new byte[0];
            FailedConnections = new List<string>();

            killers = new List<Thread>();
            spammers = new List<Thread>();
            droppers = new List<Thread>();
        }

        public void Dispose()
        {
            Thread[] kt = killers.ToArray();
            Thread[] dt = droppers.ToArray();
            Thread[] st = spammers.ToArray();

            foreach (Thread t in kt)
                t.Abort();
            foreach (Thread t in dt)
                t.Abort();
            foreach (Thread t in st)
                t.Abort();
            instance = null;
        }

        #region Methods

        /// <summary>
        /// Renames an active connection.
        /// </summary>
        /// <param name="oldname">The current name of the active connection.</param>
        /// <param name="newname">The new name of the active connection.</param>
        /// <returns>True if successful, false otherwise.</returns>
        public bool RenameActiveConnection(string oldname, string newname)
        {
            lock (ActiveConnectionsReference)
            {
                if (ActiveConnectionsReference.ContainsKey(oldname))
                {
                    if (!ActiveConnectionsReference.ContainsKey(newname))
                    {
                        ActiveConnectionsReference.Add(newname, ActiveConnectionsReference[oldname]);
                        ActiveConnectionsBackReference.Add(ActiveConnectionsReference[newname], newname);
                        ActiveConnectionsBackReference.Remove(ActiveConnectionsReference[oldname]);
                        ActiveConnectionsReference.Remove(oldname);
                        return true;
                    }
                }
            }
            return false;
        }


        /// <summary>
        /// Sets whether or not to drop connections after waiting for credentials, and if so, for how long.
        /// </summary>
        /// <param name="time">How many milliseconds to wait for credentials.</param>
        /// <param name="active">Whether to drop or not.</param>
        public void SetDropout(uint time, bool active)
        {
            dropontimeout = active;
            dropouttimer = time;
        }

        /// <summary>
        /// Set this node's credential packet.
        /// </summary>
        /// <param name="newCred">The byte array representing this node's credentials.</param>
        public void SetCredentials(byte[] newCred)
        {
            lock (credentials)
            {
                credentials = newCred;
            }
        }

        /// <summary>
        /// Set the node's Packet-Request packet.  Other nodes will expect this packet, so hopefully it's standardized.
        /// </summary>
        /// <param name="Request"></param>
        public void SetRequestPacket(byte[] Request)
        {
            lock (CredentialRequestPacket)
            {
                CredentialRequestPacket = Request;
            }
        }

        /// <summary>
        /// List all active connections.
        /// </summary>
        /// <returns>A list of all active connection strings.</returns>
        public List<string> GetAllActiveConnections()
        {
            List<String> returnme;
            lock (ActiveConnectionsReference)
            {
                returnme = new List<String>(ActiveConnectionsReference.Keys);
            }
            return returnme;
        }

        /// <summary>
        /// List all inactive connections.
        /// </summary>
        /// <returns>A Dictionary of all inactive connections along with their provided credentials.  The name is [IP]_number.</returns>
        public Dictionary<string, byte[]> GetAllInactiveConnections()
        {
            Dictionary<string, byte[]> returnme = new Dictionary<string, byte[]>();
            lock (InactiveConnections)
            {
                foreach (KeyValuePair<string, KeyValuePair<AsyncSocket, byte[]>> k in InactiveConnections)
                {
                    returnme.Add(k.Key, k.Value.Value);
                }
            }
            return returnme;
        }

        /// <summary>
        /// Activates an inactive connection.
        /// </summary>
        /// <param name="oldname">The name of the inactive connection.  If the connection is not found, it will return and do nothing.</param>
        /// <param name="newname">Its new name as an active connection.  If that name is in use, it will return and do nothing.</param>
        public void ActivateConnection(string oldname, string newname)
        {
            //Check to make sure neither list has either name, while being threadsafe.  How annoying.
            lock (InactiveConnections)
            {
                if (!InactiveConnections.ContainsKey(oldname))
                {
                    return;
                }
            }
            lock (ActiveConnectionsReference)
            {
                if (ActiveConnectionsReference.ContainsKey(newname))
                {
                    return;
                }
            }

            //Now do the actual swap.
            AsyncSocket lol;
            lock (InactiveConnections)
            {
                lol = InactiveConnections[oldname].Key;
                InactiveConnections.Remove(oldname);
            }
            lock (ActiveConnectionsReference)
            {
                ActiveConnectionsReference.Add(newname, lol);
                ActiveConnectionsBackReference.Add(lol, newname);
            }
            lock (DataReceived)
            {
                DataReceived.Add(newname, new List<byte[]>());
            }
            if (ConnectionActivated != null)
                ConnectionActivated(newname);

            //Added by nroy to allow the showing of connection activation to the user
            if (NotifyOfClientConnection != null)
                NotifyOfClientConnection(lol.RemoteEndPoint.ToString() + " connected to the server.");
        }

        /// <summary>
        /// Listens for a connection on the given port.
        /// </summary>
        /// <param name="port">The port to listen on.</param>
        /// <param name="type">If true, IPV4, if false, IPV6</param>
        public void ListenForConnection(int port, bool IPV4)
        {
            lock (ListeningForConnection)
            {
                if (ListeningForConnection.ContainsValue(port))
                {
                    return;
                }
                ProtocolType type;
                AddressFamily add;
                if (IPV4)
                {
                    type = ProtocolType.IP;
                    add = AddressFamily.InterNetwork;
                }
                else
                {
                    type = ProtocolType.IP;
                    add = AddressFamily.InterNetworkV6;
                }
                AsyncSocket socket = new AsyncSocket(add, SocketType.Stream, type);
                ListeningForConnection.Add(socket, port);
                socket.Bind(new System.Net.IPEndPoint(0, port));
                socket.Listen(10);
                socket.Accept();
                socket.AcceptedConnection += new AsyncSocket.AsyncSocketConnectEvent(SocketAcceptedConnection);

            }
        }

        /// <summary>
        /// Stops listening on a specific port.
        /// </summary>
        /// <param name="port">The port to stop listening on.</param>
        public void StopListening(int port)
        {
            lock (ListeningForConnection)
            {
                if (ListeningForConnection.ContainsValue(port))
                {
                    List<AsyncSocket> moarpain = new List<AsyncSocket>();
                    foreach (KeyValuePair<AsyncSocket, int> k in ListeningForConnection)
                    {
                        if (k.Value == port)
                        {
                            k.Key.Close();
                            moarpain.Add(k.Key);
                        }
                    }
                    ListeningForConnection.Remove(moarpain[0]);
                }
            }
        }

        /// <summary>
        /// Lists all the ports currently listening for connections.
        /// </summary>
        /// <returns>A list of integers representing active ports.</returns>
        public List<int> GetListeningPorts()
        {
            List<int> returnme = new List<int>();
            lock (ListeningForConnection)
            {
                returnme.AddRange(ListeningForConnection.Values);
            }
            return returnme;
        }

        /// <summary>
        /// Will disconnect an active connection and remove it from the pool of active connections.
        /// </summary>
        /// <param name="name">The name of the connection to kill.</param>
        public void KillActiveConnection(string name)
        {
            lock (ActiveConnectionsReference)
            {
                if (ActiveConnectionsReference.ContainsKey(name))
                {
                    ActiveConnectionsBackReference.Remove(ActiveConnectionsReference[name]);
                    AsyncSocket a = ActiveConnectionsReference[name];
                    ActiveConnectionsReference.Remove(name);
                    a.Disconnect(false);
                }
            }
            lock (DataReceived)
            {
                if (DataReceived.ContainsKey(name))
                {
                    DataReceived.Remove(name);
                }
            }
        }

        /// <summary>
        /// Will kill an inactive connection.  Use if credentials aren't valid.
        /// </summary>
        /// <param name="name">The name of the inactive connection.</param>
        public void KillInactiveConnection(string name)
        {
            lock (InactiveConnections)
            {
                if (InactiveConnections.ContainsKey(name))
                {
                    AsyncSocket a = InactiveConnections[name].Key;
                    InactiveConnections.Remove(name);
                    a.Disconnect(false);                                       
                }
            }
        }

        /// <summary>
        /// Will connect to a remote location.  Upon success, it will be an active connection.
        /// </summary>
        /// <param name="ConnectionName">The name of the connection.  Will do nothing if name is in use by an active connection.</param>
        /// <param name="add">Address Family</param>
        /// <param name="HostName">The host to connect to.</param>
        /// <param name="port">The port to connect on.</param>
        /// <param name="type">if true, IPV4.  If false, IPV6.</param>
        public void MakeConnection(string ConnectionName, string HostName, int port, bool IPV4)
        {
            lock (ActiveConnectionsReference)
            {
                if (ActiveConnectionsReference.ContainsKey(ConnectionName))
                {
                    return;
                }
            }
            lock (PendingApproval)
            {
                if (PendingApproval.ContainsValue(ConnectionName))
                {
                    return;
                }
            }
            lock (WaitingForConnection)
            {
                if (!WaitingForConnection.ContainsValue(ConnectionName))
                {
                    ProtocolType type;
                    AddressFamily add;
                    if (IPV4)
                    {
                        type = ProtocolType.IP;
                        add = AddressFamily.InterNetwork;
                    }
                    else
                    {
                        type = ProtocolType.IP;
                        add = AddressFamily.InterNetworkV6;
                    }
                    AsyncSocket socket = new AsyncSocket(add, SocketType.Stream, type);
                    socket.Connected += new AsyncSocket.AsyncSocketEvent(SocketConnected);
                    WaitingForConnection.Add(socket, ConnectionName);
                    socket.Connect(HostName, port);
                    ParameterizedThreadStart ts = new ParameterizedThreadStart(KillSwitch);
                    Thread t = new Thread(ts);
                    t.Name = ConnectionName + " Kill Switch";
                    t.Start(socket);
                    killers.Add(t);
                }
            }
        }

        /// <summary>
        /// Sends data along the given connection.
        /// </summary>
        /// <param name="Connection">An active connection.  Will do nothing if it does not exist.</param>
        /// <param name="Data">The data to send.</param>
        public void SendData(string Connection, byte[] Data)
        {
            lock (ActiveConnectionsReference)
            {
                if (!ActiveConnectionsReference.ContainsKey(Connection))
                {
                    return;
                }
                try
                {
                    ActiveConnectionsReference[Connection].Send(Data);
                }
                catch
                {
                    ActiveConnectionsBackReference.Remove(ActiveConnectionsReference[Connection]);
                    ActiveConnectionsReference.Remove(Connection);
                    lock (DroppedConnections)
                    {
                        DroppedConnections.Add(Connection);
                    }
                }
            }
        }

        /// <summary>
        /// Will return all data received on connection since last clear.
        /// </summary>
        /// <param name="Connection">The connection that has received data.</param>
        /// <param name="ClearData">Whether to wipe the data once it's been returned.</param>
        /// <returns>A list of byte[] containing all the data received.</returns>
        public List<byte[]> GetData(string Connection, bool ClearData)
        {
            List<byte[]> returnme = new List<byte[]>();
            lock (DataReceived)
            {
                if (!DataReceived.ContainsKey(Connection))
                    return returnme;
                returnme.AddRange(DataReceived[Connection]);
                if (ClearData)
                    DataReceived[Connection].Clear();
            }
            return returnme;
        }

        /// <summary>
        /// Will checks to see if a connection has been dropped.
        /// </summary>
        /// <param name="Connection">Connection dropped.</param>
        public void CheckForDisconnect(string Connection)
        {
            lock (ActiveConnectionsReference)
            {
                if (ActiveConnectionsReference.ContainsKey(Connection))
                {
                    if (!ActiveConnectionsReference[Connection].IsConnected)
                    {
                        KillActiveConnection(Connection);
                        lock (DroppedConnections)
                        {
                            DroppedConnections.Add(Connection);
                        }
                    }
                }
            }
        }

        /// <summary>
        /// Will return all data received across all connections since last clear.
        /// </summary>
        /// <param name="ClearData">Whether to wipe the data once it's been returned.</param>
        /// <returns>A list of byte[] containing all the data received.</returns>
        public List<byte[]> GetData(bool ClearData)
        {
            List<byte[]> returnme = new List<byte[]>();
            lock (DataReceived)
            {
                foreach (KeyValuePair<string, List<byte[]>> k in DataReceived)
                {
                    returnme.AddRange(k.Value);
                }
                if (ClearData)
                    foreach (string s in DataReceived.Keys)
                    {
                        DataReceived[s].Clear();
                    }
            }
            return returnme;

        }

        /// <summary>
        /// Lists all dropped active and inactive connections, as well as failed connection attempts.
        /// </summary>
        /// <param name="clear">Whether to clear the list or not once it's returned.</param>
        /// <returns>A list of all dropped connection names.</returns>
        public List<string> GetDroppedConnections(bool clear)
        {
            List<string> returnme = new List<string>();
            lock (DroppedConnections)
            {
                returnme.AddRange(DroppedConnections);
                if (clear)
                {
                    DroppedConnections.Clear();
                }
            }
            return returnme;
        }

        /// <summary>
        /// Lists all failed connection attempts.  Does not list failed listening attempts.
        /// </summary>
        /// <param name="clear">Whether to clear the list after returning.</param>
        /// <returns>A list of all failed connection names.</returns>
        public List<string> GetFailedConnections(bool clear)
        {
            List<string> returnme = new List<string>();
            lock (FailedConnections)
            {
                returnme.AddRange(FailedConnections);
                if (clear)
                {
                    FailedConnections.Clear();
                }
            }
            return returnme;

        }


        #endregion

        #region Callbacks

        private void SocketAcceptedConnection(object sender, SocketConnectEventArgs e)
        {
            AsyncSocket socket = sender as AsyncSocket;
            lock (ListeningForConnection)
            {
                if (!ListeningForConnection.ContainsKey(socket))
                {
                    return;
                }
                socket.Listen(10);
                socket.Accept();
            }
            socket = e.ConnectedClient;
            lock (AwaitingCredentials)
            {
                string name = ((System.Net.IPEndPoint)e.ConnectedClient.RemoteEndPoint).Address.ToString();
                socket.DataSent += new AsyncSocket.AsyncSocketDataEvent(SocketDataSent);
                socket.Send(CredentialRequestPacket);
                socket.DataReceived += new AsyncSocket.AsyncSocketDataEvent(SocketDataReceived);
                socket.Receive(2048, SocketFlags.None);

                ParameterizedThreadStart threadz = new ParameterizedThreadStart(SpamRequest);
                Thread tred = new Thread(threadz);
                tred.Name = name + " Request Spammer";
                AwaitingCredentials.Add(socket, name);
                tred.Start(socket);
                spammers.Add(tred);

                if (dropontimeout)
                {
                    ParameterizedThreadStart threadstart = new ParameterizedThreadStart(WaitForDrop);
                    Thread t = new Thread(threadstart);
                    t.Name = name + " Drop Watcher";
                    t.Start(socket);
                    droppers.Add(t);
                }
            }
        }

        void SocketDataSent(object sender, SocketDataEventArgs e)
        {
            return;
        }

        private void SocketConnected(object sender, SocketEventArgs e)
        {
            AsyncSocket socket = sender as AsyncSocket;
            string name;
            lock (WaitingForConnection)
            {
                if (!WaitingForConnection.ContainsKey(socket))
                {
                    return;
                }
                name = WaitingForConnection[socket];
                WaitingForConnection.Remove(socket);
            }
            lock (PendingApproval)
            {
                PendingApproval.Add(socket, name);
                socket.DataReceived += new AsyncSocket.AsyncSocketDataEvent(SocketDataReceived);
                socket.Disconnected += new AsyncSocket.AsyncSocketEvent(SocketDisconnected);
                socket.SocketError += new AsyncSocket.AsyncSocketErrorEvent(SocketSocketError);
                socket.Receive(2048, SocketFlags.None);
            }
        }

        void SocketSocketError(object sender, AsyncSocketException e)
        {
            //Possibly an error log?
            ErrorManager.getInstance().Catch(e);
        }

        void SocketDisconnected(object sender, SocketEventArgs e)
        {
            AsyncSocket socket = sender as AsyncSocket;
            string name = "";
            bool terminate = true;

            //added by nroy to notify the console on client disconnect
            if (NotifyOfClientDisconnection != null)
                NotifyOfClientDisconnection(socket.RemoteEndPoint.ToString() + " has disconnected from the server.");

            if (terminate)
            {
                lock (ActiveConnectionsReference)
                {
                    if (ActiveConnectionsBackReference.ContainsKey(socket))
                    {
                        terminate = false;
                        name = ActiveConnectionsBackReference[socket];
                        ActiveConnectionsBackReference.Remove(socket);
                        ActiveConnectionsReference.Remove(name);
                    }
                }
                if (!terminate)
                {
                    lock (DataReceived)
                    {
                        DataReceived.Remove(name);
                    }

                }
            }
            if (terminate)
            {
                lock (InactiveConnections)
                {
                    foreach (KeyValuePair<string, KeyValuePair<AsyncSocket, byte[]>> k in InactiveConnections)
                    {
                        if (k.Value.Key.Equals(socket))
                        {
                            terminate = false;
                            name = k.Key;
                        }
                    }
                    InactiveConnections.Remove(name);
                }
            }
            if (terminate)
            {
                lock (PendingApproval)
                {
                    if (PendingApproval.ContainsKey(socket))
                    {
                        terminate = false;
                        name = PendingApproval[socket];
                    }
                    PendingApproval.Remove(socket);
                }
            }
            if (!terminate)
            {
                DroppedConnections.Add(name);
            }
        }

        private void SocketDataReceived(object sender, SocketDataEventArgs e)
        {
            AsyncSocket socket = sender as AsyncSocket;
            bool unparsed = true;
            byte[] data = { };
            //Check to see if it's an accepted connection.  If so, claim the data has been parsed, and set
            //flag to write to DataReceived.  Specifically only locking one list at a time to prevent
            //deadlocks.
            if (unparsed)
            {
                lock (ActiveConnectionsReference)
                {
                    if (ActiveConnectionsBackReference.ContainsKey(socket))
                    {
                        unparsed = false;
                        data = e.Data;
                    }
                }
                //If the data has been just parsed, it was an active connection.
                if (!unparsed)
                {
                    lock (DataReceived)
                    {
                        DataReceived[ActiveConnectionsBackReference[socket]].Add(data);
                    }
                }
            }

            //If we're waiting for credentials from someone, then we'll accept them here.
            //Upon accepted credentials, connection becomes an inactive connection.
            if (unparsed)
            {
                string IP = "";
                lock (AwaitingCredentials)
                {
                    if (AwaitingCredentials.ContainsKey(socket))
                    {
                        unparsed = false;
                        IP = AwaitingCredentials[socket];
                        AwaitingCredentials.Remove(socket);
                    }
                }
                //If data has been just parsed, it was a connection awaiting credentials, and we move them to Inactive Connections.
                if (!unparsed)
                {
                    lock (InactiveConnections)
                    {
                        InactiveConnections.Add(ConnectionManager.GetNewName(IP), new KeyValuePair<AsyncSocket, byte[]>(socket, e.Data));
                        socket.Disconnected += new AsyncSocket.AsyncSocketEvent(SocketDisconnected);
                        socket.SocketError += new AsyncSocket.AsyncSocketErrorEvent(SocketSocketError);
                    }
                }
            }
            //If the socket is pending approval, we check to see if the server responds approval.  If so, we send credentials and add it to Active Connections.
            //If the connection is then refused, it's simply dropped.
            if (unparsed)
            {
                string name = "";
                lock (PendingApproval)
                {
                    if (PendingApproval.ContainsKey(socket))
                    {
                        unparsed = false;
                        name = PendingApproval[socket];
                        PendingApproval.Remove(socket);
                    }
                }
                if (!unparsed)
                {
                    if (CheckRequest(e.Data))
                    {
                        lock (ActiveConnectionsReference)
                        {
                            if (!ActiveConnectionsReference.ContainsKey(name))
                            {
                                ActiveConnectionsBackReference.Add(socket, name);
                                ActiveConnectionsReference.Add(name, socket);
                                socket.Send(credentials);
                            }
                        }
                        lock (DataReceived)
                        {
                            if(!DataReceived.ContainsKey(name))
                                DataReceived.Add(name, new List<byte[]>());
                        }
                        if (ConnectionActivated != null)
                            ConnectionActivated(name);
                    }
                    else
                    {
                        socket.Disconnect(false);
                        lock (DroppedConnections)
                        {
                            DroppedConnections.Add(name);
                        }
                    }
                }
            }
            socket.Receive(2048, SocketFlags.None);
        }

        #endregion

        #region Private Methods

        /// <summary>
        /// Will be running on its own thread.  This method waits the specified amount of time, and will drop a connection if credentials haven't been received.
        /// </summary>
        /// <param name="parameter">The AsyncSocket.</param>
        private void WaitForDrop(object parameter)
        {
            Thread.Sleep((int)dropouttimer);
            AsyncSocket socket = parameter as AsyncSocket;
            droppers.Remove(Thread.CurrentThread);
            lock (AwaitingCredentials)
            {
                if (AwaitingCredentials.ContainsKey(socket))
                {
                    AwaitingCredentials.Remove(socket);
                    socket.Disconnect(false);
                }
            }
        }

        /// <summary>
        /// Will spam requests for credentals until it's timed out, or they are received.
        /// </summary>
        /// <param name="parameter">The AsyncSocket.</param>
        private void SpamRequest(object parameter)
        {
            Thread.Sleep((int)(dropouttimer / 20));
            AsyncSocket socket = parameter as AsyncSocket;
            spammers.Remove(Thread.CurrentThread);
            bool die = false; ;
            lock (AwaitingCredentials)
            {
                if (AwaitingCredentials.ContainsKey(socket))
                {
                    try
                    {
                        socket.Send(CredentialRequestPacket);
                    }
                    catch
                    {
                        die = true;
                    }
                }
                else
                    die = true;
            }
            if (die)
                return;
            ParameterizedThreadStart ts = new ParameterizedThreadStart(SpamRequest);
            Thread t = new Thread(SpamRequest);
            t.Name = Thread.CurrentThread.Name;
            t.Start(socket);
            spammers.Add(t);
        }

        /// <summary>
        /// Will return a new temporary name.
        /// </summary>
        /// <returns>A temporary name.</returns>
        static private string GetNewName(string old)
        {
            string returnme = "";
            returnme += old;
            returnme += "_";
            returnme += log;
            ConnectionManager.log++;
            return returnme;
        }

        /// <summary>
        /// Will check byte arrays for equality even if they are different sizes.
        /// </summary>
        /// <param name="buffer">The buffer to check against Request.</param>
        /// <returns>True if they're equal, false otherwise.</returns>
        private bool CheckRequest(byte[] buffer)
        {
            if (buffer.Length < CredentialRequestPacket.Length)
                return false;
            for (int i = 0; i < CredentialRequestPacket.Length; i++)
            {
                if (buffer[i] != CredentialRequestPacket[i])
                {
                    return false;
                }
            }
            return true;

        }

        /// <summary>
        /// Will run on its own thread.  If a connection attempt goes on for too long, it kills it.
        /// </summary>
        /// <param name="parameter">The socket.</param>
        private void KillSwitch(object parameter)
        {
            Thread.Sleep((int)dropouttimer);
            AsyncSocket socket = parameter as AsyncSocket;
            string name = "";
            killers.Remove(Thread.CurrentThread);
            lock (WaitingForConnection)
            {
                if (WaitingForConnection.ContainsKey(socket))
                {
                    name = WaitingForConnection[socket];
                    socket.Close();
                    WaitingForConnection.Remove(socket);
                }
                else
                {
                    return;
                }
            }
            lock (FailedConnections)
            {
                FailedConnections.Add(name);
            }
        }

        #endregion
    }
}
