using System;
using System.Collections.Generic;
using System.Text;
using System.Reflection;

namespace MrHomer.Networking
{
    /// <summary>
    /// A packet class.  Essentially  useful way to wrap data.
    /// </summary>
    public class Packet
    {
        /// <summary>
        /// The Deserialization error delegate.
        /// </summary>
        /// <param name="data">The bytestream that was attempted to be deserialized.</param>
        public delegate void DeseralizationErrorEvent(byte[] data);
        /// <summary>
        /// Upon a deserialization error, this event will fire.
        /// </summary>
        public static event DeseralizationErrorEvent DeserializationError;

        private string sender;
        /// <summary>
        /// The packet's sender.
        /// </summary>
        public string Sender
        {
            get
            {
                return sender;
            }
            set
            {
                sender = value;
            }
        }

        private string recipient;

        /// <summary>
        /// The packet's recipient.
        /// </summary>
        public string Recipient
        {
            get
            {
                return recipient;
            }
            set
            {
                recipient = value;
            }
        }

        /// <summary>
        /// The packet's subject.
        /// </summary>
        private string subject;
        public string Subject
        {
            get
            {
                
                return subject;
            }
            set
            {
                subject = value;
            }
        }

        private string type;

        private IHSerializable body;

        /// <summary>
        /// The packet's body.
        /// </summary>
        public IHSerializable Body
        {
            get
            {
                return body;
            }
            set
            {
                body = value;
            }
        }

        /// <summary>
        /// The main constructor for creating packets.
        /// </summary>
        /// <param name="Sender">Choose a sender.</param>
        /// <param name="Recipient">Choose a recipient.</param>
        /// <param name="Subject">Choose a subject.</param>
        /// <param name="Body">And give it an IHSerializable body.</param>
        public Packet(string Sender, string Recipient, string Subject, IHSerializable Body)
        {
            sender = Sender;
            recipient = Recipient;
            subject = Subject;
            type = Body.GetType().AssemblyQualifiedName;
            body = Body;
        }

        /// <summary>
        /// Packet deserialization.
        /// </summary>
        /// <param name="data">The data to deserialize.</param>
        public Packet(byte[] data)
        {
            try
            {
                System.Text.UTF8Encoding encoder = new UTF8Encoding();
                int offset = 0; //current location

                int length = BitConverter.ToInt32(data, offset);
                offset += 4;
                sender = encoder.GetString(data, offset, length);
                offset += length;

                length = BitConverter.ToInt32(data, offset);
                offset += 4;
                recipient = encoder.GetString(data, offset, length);
                offset += length;

                length = BitConverter.ToInt32(data, offset);
                offset += 4;
                subject = encoder.GetString(data, offset, length);
                offset += length;

                length = BitConverter.ToInt32(data, offset);
                offset += 4;
                type = encoder.GetString(data, offset, length);
                offset += length;

                length = BitConverter.ToInt32(data, offset);
                offset += 4;

                byte[] stuff = new byte[length];
                for (int i = 0; i < length; i++)
                {
                    stuff[i] = data[i + offset];
                }

                object o = System.Activator.CreateInstance(Type.GetType(type));
                body = o as IHSerializable;
                body.Deserialize(stuff);
            }
            catch
            {
                if (DeserializationError != null)
                {
                    Packet.DeserializationError(data);
                }
            }
        }

        /// <summary>
        /// Serializes the packet into a byte stream.
        /// </summary>
        /// <returns>The byte stream.</returns>
        public byte[] Serialize()
        {
            System.Text.UTF8Encoding encoder = new UTF8Encoding();
            int total = 20;
            byte[] bSender = encoder.GetBytes(sender);
            byte[] bRecipient = encoder.GetBytes(recipient);
            byte[] bSubject = encoder.GetBytes(subject);
            byte[] bType = encoder.GetBytes(type);
            byte[] bBody = body.Serialize();
            byte[] lSender = BitConverter.GetBytes(bSender.Length);
            byte[] lRecipient = BitConverter.GetBytes(bRecipient.Length);
            byte[] lSubject = BitConverter.GetBytes(bSubject.Length);
            byte[] lType = BitConverter.GetBytes(bType.Length);
            byte[] lBody = BitConverter.GetBytes(bBody.Length);
            total += bSender.Length;
            total += bRecipient.Length;
            total += bSubject.Length;
            total += bBody.Length;
            total += bType.Length;

            byte[] s = new byte[total];
            int location = 0;

            lSender.CopyTo(s, location);
            location += lSender.Length;
            bSender.CopyTo(s, location);
            location += bSender.Length;
            
            lRecipient.CopyTo(s, location);
            location += lRecipient.Length;            
            bRecipient.CopyTo(s, location);
            location += bRecipient.Length;

            lSubject.CopyTo(s, location);
            location += lSubject.Length;
            bSubject.CopyTo(s, location);
            location += bSubject.Length;

            lType.CopyTo(s, location);
            location += lType.Length;
            bType.CopyTo(s, location);
            location += bType.Length;
            
            lBody.CopyTo(s, location);
            location += lBody.Length;
            bBody.CopyTo(s, location);
            location += bBody.Length;

            return s;
        }
    }
}
