using System;
using System.Collections.Generic;
using System.Text;
using System.IO;
using System.Reflection;
using MrHomer.Plugins;

namespace MrHomer.Plugins
{
    /// <summary>
    /// The plugin manager for MrHomer.  And now, a poem.
    /// 
    /// Once upon a midnight dreary, while I programmed weak and weary,
    /// Over many a lame and horrendous volume of programming lore,
    /// While I coded, quickly typing, suddenly there came a griping,
    /// Of the compiler whining, whining warnings, I ignore.
    /// T'is not possible, I muttered, seeing the warnings, I ignore -
    /// For my code is so hardcore.
    /// </summary>
    public class PluginManager
    {

        #region Singleton
        static private PluginManager instance;
        /// <summary>
        /// Singleton Plugin method.  Will return the current instance of the PluginManager.
        /// </summary>
        /// <returns>The current PluginManager instance.</returns>
        static public PluginManager getInstance()
        {
            if (instance == null)
                instance = new PluginManager();
            return instance;
        }
        #endregion

        #region Private Fields
        private Dictionary<Type, Dictionary<string, Type>> types;
        private List<Type> interfaces;
        #endregion

        #region Attribute Enforcement

        /// <summary>
        /// Given an interface attribute name, this method will go through the interface's attribute and pull aside
        /// all the attributes with that name, and store their values.  Then it will check to make sure the plugin's
        /// attributes contains one of each of the interface attribute's value.  For example:
        /// IMyPlugin has:  [MrHomerPlugin("required", "Name")]
        ///                 [MrHomerPlugin("required", "Description")]
        /// MyPlugin has:   [MrHomerPlugin("Name", "My Plugin")]
        /// And you call:   EnforceAttributes("required", typeof(MyPlugin))
        /// It will return false because MyPlugin doesn't have [MrHomerPlugin("Description", "")]
        /// </summary>
        /// <param name="InterfaceAttributeName">The interface attribute name.</param>
        /// <param name="plugin">The plugin to check.</param>
        /// <returns>True if all interface attributes specified have a plugin version, false if otherwise.</returns>
        public bool EnforceAttributes(string InterfaceAttributeName, Type plugin)
        {
            List<string> iatt = new List<string>();
            foreach (Type t in plugin.GetInterfaces())
            {
                if (interfaces.Contains(t))
                {
                    foreach (MrHomerPluginAttribute mhpa in t.GetCustomAttributes(false))
                    {
                        if (mhpa.Name.Equals(InterfaceAttributeName))
                        {
                            if(!iatt.Contains(mhpa.Value))
                                iatt.Add(mhpa.Value);
                        }
                    }
                }
            }

            foreach (MrHomerPluginAttribute t in plugin.GetCustomAttributes(false))
            {
                if (iatt.Contains(t.Name))
                    iatt.Remove(t.Name);
            }
            return (iatt.Count == 0);
        }

        #endregion

        #region Plugin Object Creation

        /// <summary>
        /// Returns an instance of the given plugin type.
        /// </summary>
        /// <typeparam name="E">The plugin's interface.</typeparam>
        /// <param name="plugin">The plugin's type.</param>
        /// <param name="args">The arguments.</param>
        /// <returns>Will return null if the generic type is not the plugin's interface or if the arguments are incorrect.</returns>
        public E CreateInstance<E>(Type plugin, object[] args)
        {
            E myObject = default(E);
            try
            {
                myObject = (E)System.Activator.CreateInstance(plugin, args);
            }
            catch
            { }
            return myObject;
        }

        /// <summary>
        /// Returns an instance of the given plugin type.
        /// </summary>
        /// <param name="plugin">The plugin's type.</param>
        /// <param name="args">The arguments.</param>
        /// <returns>Will return null if the arguments are incorrect.</returns>
        public object CreateInstance(Type plugin, object[] args)
        {
            object myObject = null;
            try
            {
                myObject = System.Activator.CreateInstance(plugin, args);
            }
            catch
            { }
            return myObject;
        }

        #endregion

        #region Plugin Type Accessors
        /// <summary>
        /// Used to get a list of available plugin Types.
        /// </summary>
        /// <returns>A list of available plugin types.</returns>
        public List<Type> getInterfaces()
        {
            return new List<Type>(interfaces);
        }

        /// <summary>
        /// Gets a list of all the MrHomerPluginAttributes that the type currently contains.
        /// </summary>
        /// <param name="plugin">The plugin type.</param>
        /// <returns>A list of attributes.</returns>
        public List<MrHomerPluginAttribute> getAttributes(Type plugin)
        {
            List<MrHomerPluginAttribute> returnme = new List<MrHomerPluginAttribute>();
            foreach(MrHomerPluginAttribute t in plugin.GetCustomAttributes(false))
            {
                returnme.Add(t);
            }
            return returnme;
        }

        /// <summary>
        /// Used to get a list of available plugins given a plugin type.
        /// </summary>
        /// <param name="iface">The type of plugins you wish to get.</param>
        /// <returns>A list of available plugins.</returns>
        public List<Type> getPlugins(Type iface)
        {
            return new List<Type>(types[iface].Values);
        }
        #endregion

        #region Constructor
        /// <summary>
        /// Oh snap, private!!
        /// </summary>
        private PluginManager()
        {
            interfaces = new List<Type>();
            types = new Dictionary<Type, Dictionary<string, Type>>();
            Assembly me = Assembly.GetExecutingAssembly();
        }
        #endregion

        #region Plugin Interface and Class Loading

        /// <summary>
        /// Loads interfaces for plugin basics from the current assembly.
        /// </summary>
        public void LoadInterfaceTypes(Assembly asm)
        {

            if (asm == null)
            {
                throw new Exception("Can't load from a null assembly.");
            }

            foreach (Type type in asm.GetTypes())
            {
                if (type.IsInterface)
                {
                    foreach (Type typeface in type.GetInterfaces())
                    {
                        if (typeface.Equals(typeof(IPlugin)))
                        {
                            if (!interfaces.Contains(type))
                                interfaces.Add(type);
                        }
                    }
                }
            }

            foreach (Type type in interfaces)
            {
                if (!types.ContainsKey(type))
                {
                    types.Add(type, new Dictionary<string, Type>());
                }
            }
        }

        /// <summary>
        /// Loads interfaces for plugin basics from an outside DLL.
        /// </summary>
        /// <param name="path">Path to the DLL.  Better give it a real path, or you're taking a one-way train to Exception town.</param>
        public void LoadInterfaceTypes(string path)
        {
           if (!File.Exists(path))
            {
                throw new Exception("Plugin Manager: Dynamic Link Library '" + path + "' not found.");
            }

            Assembly asm;
            
            asm = Assembly.LoadFile(path);
            if(asm == null)
            {
                throw new Exception("Plugin manager: Dynamic Link Library '" + path + "' cannot be loaded.");
            }

            foreach (Type type in asm.GetTypes())
            {
                if (type.IsInterface)
                {
                    foreach (Type typeface in type.GetInterfaces())
                    {
                        if (typeface.Equals(typeof(IPlugin)))
                        {
                            if(!interfaces.Contains(type))
                                interfaces.Add(type);
                        }
                    }
                }
            }

            foreach (Type type in interfaces)
            {
                if (!types.ContainsKey(type))
                {
                    types.Add(type, new Dictionary<string, Type>());
                }
            }
        }

        /// <summary>
        /// Will load plugins given a DLL.  Will whine incessantly and throw exceptions if you give it a false DLL.
        /// </summary>
        /// <param name="path">The path to the DLL.</param>
        public void LoadPluginTypes(string path)
        {
            if (!File.Exists(path))
            {
                throw new Exception("Plugin Manager: Dynamic Link Library '" + path + "' not found.");
            }

            Assembly asm;
            
            asm = Assembly.LoadFile(path);

            LoadPluginTypes(asm);
        }

        public void LoadPluginTypes(Assembly asm)
        {
            if (asm == null)
            {
                throw new Exception("Can't load from a null assembly.");
            }
            
            foreach (Type type in asm.GetTypes())
            {
                if (type.IsClass)
                {
                    foreach (Type iface in interfaces)
                    {
                        foreach (Type typeface in type.GetInterfaces())
                        {
                            if (typeface.Equals(iface) && !types[iface].ContainsKey(type.AssemblyQualifiedName))
                                types[iface].Add(type.AssemblyQualifiedName, type);
                        }
                    }
                }
            }
        }
        #endregion
    }
}
