using System;
using System.Collections.Generic;
using System.Text;
using AsyncSockets;
using System.Net.Sockets;

namespace MrHomer.Security
{
    public class SecurityManager : ICryptographer, IAuthenticator, IKeyManager
    {
        private IAuthenticator auth;
        private ICryptographer crypto;
        private IKeyManager keys;

        #region Singleton Methods

        private static SecurityManager instance;

        public static SecurityManager getInstance()
        {
            if (instance == null)
                throw new Exception("Security Manager not initialized.");
            return instance;
        }

        public static SecurityManager getInstance(IAuthenticator Auth, ICryptographer Crypto, IKeyManager KeyManager)
        {
            if (instance == null)
                instance = new SecurityManager(Auth, Crypto, KeyManager);
            return instance;
        }

        #endregion

        private SecurityManager(IAuthenticator Auth, ICryptographer Crypto, IKeyManager KeyManager)
        {
            auth = Auth;
            crypto = Crypto;
            keys = KeyManager;
        }

        public void Reconfigure(IAuthenticator Auth, ICryptographer Crypto, IKeyManager KeyManager)
        {
            auth = Auth;
            crypto = Crypto;
            keys = KeyManager;
        }

        #region ICryptographer Members

        /// <summary>
        /// Encrypts a message with the active encryption system
        /// </summary>
        /// <param name="message">The plaintext to encrypt</param>
        /// <param name="Key">The encryption key</param>
        /// <param name="Details">Additional information for the Cryptographer</param>
        /// <returns>The cyphertext</returns>
        public byte[] Encrypt(byte[] message, byte[] Key, object[] Details)
        {
            return crypto.Encrypt(message, Key, Details);
        }

        /// <summary>
        /// Decrypts a message with the active encryption system
        /// </summary>
        /// <param name="message">The cyphertext to decrypt</param>
        /// <param name="Key">The encryption key</param>
        /// <param name="Details">Additional information for the Cryptographer</param>
        /// <returns>The plaintext</returns>
        public byte[] Decrypt(byte[] message, byte[] Key, object[] Details)
        {
            return crypto.Decrypt(message, Key, Details);
        }

        #endregion

        #region IAuthenticator Members

        /// <summary>
        /// Checks to see if an object is registered
        /// </summary>
        /// <param name="Name">The object to check</param>
        /// <param name="Type">The type of object</param>
        /// <param name="Details">Additional information for the Authenticator</param>
        /// <returns>A boolean indicating the system registration status</returns>
        public bool IsRegistered(string Name, string Type, object[] Details)
        {
            return auth.IsRegistered(Name, Type, Details);
        }

        /// <summary>
        /// Authenticates an object
        /// </summary>
        /// <param name="Name">The object to authenticate</param>
        /// <param name="Type">The type of object</param>
        /// <param name="Details">Additional information about the object</param>
        /// <returns>The object's authentication status</returns>
        public bool Authenticate(string Name, string Type, object[] Details)
        {
            return auth.Authenticate(Name, Type, Details);
        }

        #endregion

        #region IKeyManager Members

        /// <summary>
        /// Retreives a key from the system
        /// </summary>
        /// <param name="Name">The key name to retreive</param>
        /// <param name="Details">Additional information about the key</param>
        /// <returns>The key</returns>
        public byte[] GetKey(string Name, object[] Details)
        {
            return keys.GetKey(Name, Details);
        }

        /// <summary>
        /// Sets the key stored in the system
        /// </summary>
        /// <param name="Name">The key name to set</param>
        /// <param name="Key">The new key</param>
        /// <param name="Details">Additional information about the key</param>
        public void SetKey(string Name, byte[] Key, object[] Details)
        {
            keys.SetKey(Name, Key, Details);
        }

        /// <summary>
        /// Determines if the the requested key is available
        /// </summary>
        /// <param name="Name">The name of the key</param>
        /// <param name="Details">Additional information about the object</param>
        /// <returns>Whether the key is in the key manager</returns>
        public bool ContainsKey(string Name, object[] Details)
        {
            return keys.ContainsKey(Name, Details);
        }

        /// <summary>
        /// Removes the key for they given object
        /// </summary>
        /// <param name="Name">The key name to remove</param>
        /// <param name="Details">Additional information about the key</param>
        public void RemoveKey(string Name, object[] Details)
        {
            keys.RemoveKey(Name, Details);
        }

        #endregion
    }
}
