using System;
using System.Collections.Generic;
using System.Text;
using DuffBeer.Sensors.Interfaces;
using DuffBeer.Sensors.Utility;
using MrHomer.Plugins;
using EnterpriseDT.Net.Ftp;
using System.IO;
using DuffBeer;
using MrHomer;
using DuffBeer.Database.StructureTypes;

namespace On_Site_Server
{
    [MrHomerPlugin("Name", "PV Sensor")]
    [MrHomerPlugin("Make", "MeteoControl")]
    [MrHomerPlugin("Model", "LIGHT+")]
    [MrHomerPlugin("Description", "Sensor Monitoring the PV Panels")]
    class PVSensor : ISensor
    {
        #region ISensor Members

        private string name;
        FTPConnection ftp;
        private string datafile;
        private bool loaded;
        public PVSensor()
        {
             ftp = new FTPConnection();                        
             loaded = false;
        }

        public object RequestData()
        {
            requestDataFileFromSensor();        //transfer the data file over FTP connection
            return getLatestReading();          //return the latest reading from the file
        }

        /// <summary>
        /// Requests the datafile from the meteocontrol via FTP
        /// </summary>
        /// <returns>true if transfer was successful, false if not successful</returns>
        private void requestDataFileFromSensor()
        {
            /*
             * All of these lines are to be able to navigate to the correct files for today 
             */
            //remove the 20 from 2008
            string year = DateTime.Today.Year.ToString().Remove(0, 2);

            //if todays date is < 10, add a 0 in front of the day string
            string day = DateTime.Today.Day < 10 ? "0" + DateTime.Today.Day : DateTime.Today.Day.ToString();

            //if this month is < 10, add a 0 in front of the month string
            string month = DateTime.Today.Month < 10 ? "0" + DateTime.Today.Month : DateTime.Today.Month.ToString() ;
            
            //The path to the data file we are looking for
            string path = "data/" + year + month + day + "/";

            //the datafile we are looking for
            string file =  datafile + year + month + day + ".txt";

            //connect and get the info
            try
            {
                if (!ftp.IsConnected)
                    ftp.Connect();

                //ftp.Login();
                ftp.ConnectMode = FTPConnectMode.PASV;
                ftp.TransferType = FTPTransferType.ASCII;
                //ftp.ChangeWorkingDirectory(path);
                ftp.DownloadFile(name + "DataLog.txt", path + file);
                //ftp.Close();
            }
            catch (Exception e)
            {
                ErrorManager.getInstance().Catch(e);
                if (ftp.IsConnected)
                    ftp.Close();
            }
        }

        /// <summary>
        /// Gets the latest data reading from the logfile retrieved from the meteocontrol unit
        /// </summary>
        /// <returns>A datareading object with the most recent reading in it</returns>
        private DataReading getLatestReading()
        {
            try
            {
                FileStream file = new FileStream(name + "DataLog.txt", FileMode.Open, FileAccess.Read);
                StreamReader sr = new StreamReader(file);
                string s;

                do
                {
                    s = sr.ReadLine();
                }
                while (sr.Peek() != -1);

                sr.Close();
                file.Close();

                /*
                 *  Split the string at the semi-colons, this will break the data into 
                 *  Date: HH:MM:SS
                 *  Duration: int
                 *  Reading: double
                 */
                string[] values = s.Split(';');

                /*
                 *  Now break the the date part of the last string in the file into
                 *  Hour: int
                 *  Minute: int
                 *  Second: int
                 */
                string[] t = values[0].Split(':');

                /*
                 *  Build a timestamp for this data.
                 *  This assumes that the readings are only from today.
                 * 
                 */
                DateTime time = new DateTime(DateTime.Today.Year, DateTime.Today.Month,
                    DateTime.Today.Day, int.Parse(t[0]) == 24 ? 0 : int.Parse(t[0]), int.Parse(t[1]),
                    int.Parse(t[2]));

                return new DataReading(name, Double.Parse(values[2]), time.ToString());
            }
            catch (Exception exception)
            {
                SensorsTableEntry s = SensorsTableEntry.Find(SensorsTableEntry.NameColumn, name);
                SnapshotsTableEntry e = SnapshotsTableEntry.Find(SnapshotsTableEntry.SensorColumn, s.ID.ToString());

                if (e != null)
                    return new DataReading(name, e.Average, e.Timestamp);
                else
                    return new DataReading(name, 0, DateTime.Now.ToString());
            }
        }

        /// <summary>
        /// Retrieves the Config for use in the retrieving of the sensor's data file
        /// </summary>
        private void loadConfigFromBrewery()
        {
            if (!loaded)
            {
                if (ftp == null)
                    ftp = new FTPConnection();

                ftp.Timeout = 10000;
                ftp.ServerAddress = Brewery.getInstance().Config(name + "_Address");
                ftp.ServerPort = int.Parse(Brewery.getInstance().Config(name + "_Port"));
                ftp.UserName = Brewery.getInstance().Config(name + "_Username");
                ftp.Password = Brewery.getInstance().Config(name + "_Password");
                datafile = Brewery.getInstance().Config(name + "_Filename");
                loaded = true;
                try
                {
                    ftp.Connect();
                }
                catch (Exception)
                {
                    //do nothing
                }
            }
        }

        public string GetUnits()
        {
            return "kWh";
        }

        public void SetName(string s)
        {
           name = s;
           loadConfigFromBrewery();
        }

        public byte[] SerializeConfiguration()
        {
            return (new UTF8Encoding()).GetBytes(name);
        }

        public void DeserializeConfiguration(byte[] data)
        {
            name = (new UTF8Encoding()).GetString(data);
        }

        public Type ReturnType
        {
            get { return typeof(DataReading); }
        }

        #endregion

        #region IPlugin Members

        public string Author
        {
            get { return "Nathan Roy"; }
        }

        public string Version
        {
            get { return "V 0.1"; }
        }

        public string Label
        {
            get { return name; }
        }

        #endregion
    }
}
