using System;
using System.Collections.Generic;
using System.Text;
using DuffBeer.Communication;
using MrHomer;
using MrHomer.Networking;
using DuffBeer.Sensors;
using System.Threading;
using MrHomer.Security;
using DuffBeer;
using DuffBeerDriver.FakePlugins;
using DuffBeer.Security.Cryptography;
using DuffBeer.Sensors.Utility;
using DuffBeer.Messages.MessageTypes;
using MrHomer.Plugins;
using System.Timers;

namespace Remote_Requesting_Application
{
    class Core
    {
        #region Variables
        //private static string Connection = "Test Connection";
        private static volatile CommunicationManager cm;
        private static ErrorManager em;
        private static ConnectionManager mm;
        private static SensorManager sm;
        public static MainForm.DelayedMessage Logger;
        //private System.Timers.Timer connectionTimer;
        private static Thread t;
        private MainForm form;
        #endregion

        public Core(MainForm form)
        {
            this.form = form;
        }
        #region Helpers
        private static void Log(string message)
        {
            Logger(message);
        }

        private static void ErrorLog(Exception e)
        {
            Log(e.ToString() + "\n");
        }

        private static void MsgThreader()
        {
            while (cm.HasEvents)
            {
                try
                {
                    cm.DoEvents();
                    em.DumpErrors(new ErrorManager.ErrorPrinter(Core.ErrorLog));
                    Thread.Sleep(100);
                }
                catch (ThreadAbortException)
                {
                    return;
                }
            }
            Log("Event Loop Finished\n");
        }
        #endregion
        #region Methods
        /// <summary>
        /// Runs the Application
        /// </summary>
        public void Run()
        {
            Log("Welcome to the MrHomer On-site Server Program");
            performInit();
            Log("Initialization Tasks Complete...");
            Log("Running on Port: " + Brewery.getInstance().Config("Port"));
            Log("Click Connect to open connection with the remote computer...\n");
        }

        /// <summary>
        /// Handles initialization tasks needed before execution of the program
        /// </summary>
        private void performInit()
        {
            SecurityManager.getInstance(new SuperAuth(), new NRotN_Crypto(), new SuperKey());
            Brewery.getInstance();
            cm = CommunicationManager.getInstance();
            mm = ConnectionManager.getInstance();
            em = ErrorManager.getInstance();
            sm = SensorManager.getInstance();

            t = new Thread(new ThreadStart(Core.MsgThreader));
            t.Name = "Events Loop";
            t.Start();
        }

        /// <summary>
        /// Quits the program
        /// </summary>
        public void CleanUpAndQuit()
        {
            if (mm.GetAllActiveConnections().Count > 0)
            {
                ITask disconn = new DisconnectionTask();
                disconn.InitializeTask(form.Connection, new object[] { int.Parse(Brewery.getInstance().Config("Port")), true });
                cm.AddTask(disconn);

                while (mm.GetAllActiveConnections().Count > 0) ;
                    Log("Now Disconnected!\n");
            }

            Brewery.getInstance().Dispose();
            if (t != null)
                t.Abort();
            System.Environment.Exit(1);
        }
        /// <summary>
        /// Connects to the on-site server computer
        /// </summary>
        public bool Connect(string destination)
        {
            ITask conn = new ConnectionTask();
            conn.InitializeTask(form.Connection, new object[] { destination, int.Parse(Brewery.getInstance().Config("ServerPort")), true });
            cm.AddTask(conn);

            Thread.Sleep(2000);
            if (mm.GetAllActiveConnections().Count != 0)
            {
                Log("Connected!\n");
                return true;
            }
            else
            {
                Log("Could not connect to remote computer, connection timed out.\n");
                return false;
            }
        }


        /// <summary>
        /// Disconnects from the on-site server computer
        /// </summary>
        public void Disconnect()
        {
            ITask disconn = new DisconnectionTask();
            disconn.InitializeTask(form.Connection, new object[] { int.Parse(Brewery.getInstance().Config("Port")), true });
            cm.AddTask(disconn);

            while (mm.GetAllActiveConnections().Count > 0) ;
            Log("Now Disconnected!\n");

            Thread.Sleep(0);
        }

        /// <summary>
        /// Requests Sensor data
        /// </summary>
        /// <returns>The latest sensor data</returns>
        public Dictionary<string, string> RequestSensorData(string sensor_name)
        {
           //string sensor_name = "Sensor1";
            string dataRequestVar = form.Connection + "_" + sensor_name + "_Data";

            string subject = "";
            foreach (MrHomer.Plugins.MrHomerPluginAttribute n in MrHomer.Plugins.PluginManager.getInstance().getAttributes(typeof(SensorDataRequestMessageHandler)))
            {
                if (n.Name == "Subject")
                    subject = n.Value;
            }

            Brewery.getInstance().AddVariable(dataRequestVar);

            Log("Sending the Sensor data request to the server...");
            Packet p = new Packet(Brewery.getInstance().UPRN, form.Connection, subject, new StringData(sensor_name));
            ITask task = new SendDataTask();
            task.InitializeTask(form.Connection, new object[] { p, SecurityManager.getInstance().GetKey(form.Connection, new object[] { }), new object[] { } });
            CommunicationManager.getInstance().AddTask(task);

            Log("Awaiting Response from server...");
            while (!Brewery.getInstance().IsVariableSet(dataRequestVar))
                Thread.Sleep(500);

            object o = Brewery.getInstance().PeekVariable(dataRequestVar);

            return (Dictionary<string, string>)o;
        }

        /// <summary>
        /// Requests a list of currently installed sensors from the on-site server
        /// </summary>
        public List<String> RequestSensorList()
        {
            string varName = form.Connection + "_SensorList";

            string subject = "";
            foreach (MrHomerPluginAttribute n in PluginManager.getInstance().getAttributes(typeof(SensorListNumberRequestMessageHandler)))
            {
                if (n.Name == "Subject")
                    subject = n.Value;
            }

            Brewery.getInstance().AddVariable(varName);

            
            Packet p = new Packet(Brewery.getInstance().UPRN, form.Connection, subject, new StringData(""));
            ITask task = new SendDataTask();
            task.InitializeTask(form.Connection, new object[] { p, SecurityManager.getInstance().GetKey(form.Connection, new object[] { }), new object[] { } });
            CommunicationManager.getInstance().AddTask(task);

            Log("Awaiting Response from server...");
            while (!Brewery.getInstance().IsVariableSet(varName))
                Thread.Sleep(500);

            object o = Brewery.getInstance().PeekVariable(varName);
            
            Log("Sensor list received\n");

            return (List<String>)o;
        }

        #endregion
        
    }
}
